# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/11/2020"


from silx.gui import qt
from orangewidget import gui

import tomwer.core.process.control.volumesymlink
from tomwer.web.client import OWClient
from orangewidget.settings import Setting
from orangewidget.widget import Output, Input
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.process.control.volumesymlink import (
    OutputType,
    VolumeSymbolicLinkProcess,
)
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from processview.core.manager import ProcessManager, DatasetState
import logging

_logger = logging.getLogger(__name__)


class VolumeSymLinkOW(SuperviseOW, OWClient):
    """
    link volume reconstructed at proposal file if possible else under
    the given folder
    """

    name = "volume symbolic link"
    id = "orange.widgets.tomwer.control.VolumeSymLinkOW.VolumeSymLinkOW"
    description = (
        "create a symbolic link for the latest volume reconstructed on a "
        "given folder"
    )
    icon = "icons/volumesymlink.svg"
    priority = 200
    keywords = [
        "hdf5",
        "tomwer",
        "folder",
        "volume",
        "symlink",
        "tomography",
        "symbolic link",
    ]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    ewokstaskclass = tomwer.core.process.control.volumesymlink.VolumeSymbolicLinkProcess

    _scans = Setting(list())

    class Inputs:
        data = Input(name="data", type=TomwerScanBase, doc="scan to be process")

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

    def __init__(self, parent=None):
        SuperviseOW.__init__(self, parent)
        OWClient.__init__(self)
        self._mainWidget = _OutputDir(parent=self)
        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self._mainWidget)

    @Inputs.data
    def process(self, scan):
        if scan is None:
            return
        symlink_process = VolumeSymbolicLinkProcess()
        symlink_process.set_configuration(
            {
                "output_type": self._mainWidget.getOutputType(),
                "output_folder": self._mainWidget.getOutputFolder(),
            }
        )
        ProcessManager().notify_dataset_state(
            dataset=self.scan,
            process=self,
            state=DatasetState.ON_GOING,
        )
        try:
            scan = symlink_process.process(scan=scan)
        except Exception:
            mess = "Failed to create symbolic link for {}".format(str(scan))
            _logger.processFailed(mess)
            ProcessManager().notify_dataset_state(
                dataset=self.scan, process=self, state=DatasetState.FAILED, details=mess
            )
        else:
            mess = "Create symbolic link for {}".format(str(scan))
            _logger.processSucceed(mess)
            ProcessManager().notify_dataset_state(
                dataset=self.scan,
                process=self,
                state=DatasetState.SUCCEED,
                details=mess,
            )
        self.Outputs.data.send(scan)

    def reprocess(self, dataset):
        self.process(dataset)


class _OutputDir(qt.QGroupBox):
    def __init__(self, parent):
        qt.QGroupBox.__init__(self, parent)
        self._button_grp = qt.QButtonGroup(self)
        self._button_grp.setExclusive(True)
        self.setLayout(qt.QGridLayout())
        self._oneLevelCB = qt.QRadioButton(OutputType.ONE_LEVEL_UPPER.value, self)
        self._button_grp.addButton(self._oneLevelCB)
        self.layout().addWidget(self._oneLevelCB, 1, 0, 1, 3)
        self._staticOpt = qt.QRadioButton("", self)
        self._button_grp.addButton(self._staticOpt)
        self.layout().addWidget(self._staticOpt, 2, 0, 1, 1)
        self._qleFolder = qt.QLineEdit("", self)
        self.layout().addWidget(self._qleFolder, 2, 1, 1, 1)
        self._selectPB = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectPB, 2, 2, 1, 1)

        # connect signal / slot
        self._selectPB.released.connect(self._selectFolder)
        self._qleFolder.editingFinished.connect(self._activeStaticOpt)

        # set up
        self._oneLevelCB.setChecked(True)

    def _activeStaticOpt(self, *args, **kwargs):
        self._staticOpt.setChecked(True)

    def _selectFolder(self):
        dialog = qt.QFileDialog(self)
        dialog.setWindowTitle("Select destination folder for symbolic link")
        dialog.setModal(1)
        dialog.setFileMode(qt.QFileDialog.DirectoryOnly)

        if not dialog.exec_():
            dialog.close()
            return None

        dest_folder = dialog.selectedFiles()[0]
        self._qleFolder.setText(dest_folder)

    def getOutputType(self) -> OutputType:
        if self._oneLevelCB.isChecked():
            return OutputType.ONE_LEVEL_UPPER
        elif self._staticOpt.isChecked():
            return OutputType.STATIC
        else:
            raise TypeError("Not managed")

    def getOutputFolder(self):
        return self._qleFolder.text()



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""contain the AxisProcess
"""

__authors__ = ["C.Nemoz", "H.Payno"]
__license__ = "MIT"
__date__ = "19/03/2019"

from .params import (
    AxisRP,
    DEFAULT_CMP_TAKE_LOG,
    DEFAULT_CMP_OVERSAMPLING,
    DEFAULT_CMP_SUBSAMPLING_Y,
    DEFAULT_CMP_THETA,
)
from .mode import AxisMode
from .projectiontype import ProjectionType
from .anglemode import CorAngleMode
from tomwer.core.process.task import Task
from processview.core.superviseprocess import SuperviseProcess
from tomwer.core.scan.scanbase import TomwerScanBase
from processview.core.manager import ProcessManager
from processview.core.manager import DatasetState
from tomwer.core.scan.scanfactory import ScanFactory
import logging
from tomwer.core.utils import logconfig
import numpy
from typing import Union
from tomwer.utils import docstring
from tomwer.core.utils import image
import tomwer.version
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.process.reconstruction.nabu.utils import check_sinogram_half
from nabu.estimation.cor import (
    CenterOfRotation,
    CenterOfRotationAdaptiveSearch,
    CenterOfRotationGrowingWindow,
    CenterOfRotationSlidingWindow,
)
from nabu.pipeline.estimators import SinoCORFinder

try:
    from nabu.pipeline.estimators import CompositeCOREstimator
except ImportError:
    has_composite_cor_finder = False
else:
    has_composite_cor_finder = True
from silx.io.url import DataUrl
from tomoscan.io import HDF5File
from silx.io.utils import h5py_read_dataset
import nabu

_logger = logging.getLogger(__name__)
if not has_composite_cor_finder:
    _logger.warning("No composite cor finder found at nabu level")
# vertically, work on a window having only a percentage of the frame.
pc_height = 10.0 / 100.0
# horizontally. Global method supposes the COR is more or less in center
# % of the detector:

pc_width = 50.0 / 100.0


def adapt_tomwer_scan_to_nabu(scan: TomwerScanBase):
    """simple util to convert tomwer scan to a nabu DataAnalizer and
    updating infos regarding flat and dark if needed
    """
    dataset_infos = scan.to_nabu_dataset_analyser()
    if nabu.version > "2021.1.0" and isinstance(scan, HDF5TomoScan):
        # since this version nabu does not the load of dark / flat mean/median
        # directly in the function but outside. So we need an extra call to
        # update_dataset_info_flats_darks
        from nabu.resources.nxflatfield import update_dataset_info_flats_darks

        update_dataset_info_flats_darks(
            dataset_infos,
            processes_file=scan.get_relative_file(
                "nabu_processes.h5", with_dataset_prefix=True
            ),
            flatfield_mode=None,
        )
    return dataset_infos


def compute_cor_nabu_growing_window(
    radio_1: numpy.ndarray,
    radio_2: numpy.ndarray,
    side: str,
    padding_mode,
    flip_frame_2_lr=True,
    horz_fft_width=False,
):
    """
    Call nabu.preproc.alignement.CenterOfRotationGrowingWindow.find_shift

    :param radio_1:
    :param radio_2:
    :param padding_mode: padding mode
    :param str side: side of the cor
    :param bool flip_frame_2_lr: if True will left-right flip the second frame
    :param horz_fft_width:

    :return:
    """
    nabu_class = CenterOfRotationGrowingWindow(horz_fft_width=horz_fft_width)
    # value return is relative
    return nabu_class.find_shift(
        img_1=radio_1,
        img_2=numpy.fliplr(radio_2) if flip_frame_2_lr else radio_2,
        side=side,
        roi_yxhw=None,
        padding_mode=padding_mode,
        median_filt_shape=None,
    )


def compute_cor_nabu_growing_window_sinogram(
    scan: TomwerScanBase,
):
    """
    Call nabu.preproc.alignement.CenterOfRotationGrowingWindow.find_shift

    :param TomwerScanBase scan:
    :raises: ValueError if half acquisition with an odd number of projections

    :return:
    """
    check_sinogram_half(scan)
    corfinder = SinoCORFinder(
        adapt_tomwer_scan_to_nabu(scan),
        slice_idx=scan.axis_params.sinogram_line,
        subsampling=scan.axis_params.sinogram_subsampling,
        do_flatfield=True,
        cor_options=scan.axis_params.get_nabu_cor_options(),
        logger=_logger,
    )
    res = corfinder.find_cor(method=AxisMode.growing_window.value)
    # value return is absolute
    if scan.dim_1 is None:
        _logger.warning("unagle to find image width. Set half width to 1024")
        res = res - 1024
    else:
        res = res - scan.dim_1 // 2
    return res


def compute_scan_sino_coarse_to_fine(scan):
    """
    Compute center of rotation from `sino-coarse-to-fine` algorithm for given
    scan
    :param scan:
    :raises: ValueError if half acquisition with an odd number of projections
    :return:
    """
    check_sinogram_half(scan)
    corfinder = SinoCORFinder(
        adapt_tomwer_scan_to_nabu(scan),
        slice_idx=scan.axis_params.sinogram_line,
        subsampling=scan.axis_params.sinogram_subsampling,
        do_flatfield=True,
        cor_options=scan.axis_params.get_nabu_cor_options(),
        logger=_logger,
    )
    res = corfinder.find_cor(method=AxisMode.sino_coarse_to_fine.value)
    # value return is absolute
    if scan.dim_1 is None:
        _logger.warning("unagle to find image width. Set half width to 1024")
        res = res - 1024
    else:
        res = res - scan.dim_1 // 2
    return res


def compute_scan_composite_coarse_to_fine(scan: TomwerScanBase):
    """
    Compute center of rotation from `sino-coarse-to-fine` algorithm for given
    scan
    :param scan:
    :raises: ValueError if half acquisition with an odd number of projections
    :return:
    """
    if not has_composite_cor_finder:
        _logger.error("unable to find nabu CompositeCOREstimator")
        return None

    check_sinogram_half(scan)
    theta, subsampling_y, oversampling, take_log = get_composite_options(scan)
    corfinder = CompositeCOREstimator(
        adapt_tomwer_scan_to_nabu(scan),
        theta=theta,
        subsampling_y=subsampling_y,
        oversampling=oversampling,
        cor_options=scan.axis_params.get_nabu_cor_options(),
        logger=_logger,
        take_log=take_log,
    )
    res = corfinder.find_cor(method=AxisMode.composite_coarse_to_fine.value)
    # value return is absolute
    if scan.dim_1 is None:
        _logger.warning("unagle to find image width. Set half width to 1024")
        res = res - 1024
    else:
        res = res - scan.dim_1 // 2
    return res


def get_composite_options(scan):
    theta = scan.axis_params.composite_options.get("theta", DEFAULT_CMP_THETA)
    subsampling_y = scan.axis_params.composite_options.get(
        "subsampling_y", DEFAULT_CMP_SUBSAMPLING_Y
    )
    oversampling = scan.axis_params.composite_options.get(
        "oversampling", DEFAULT_CMP_OVERSAMPLING
    )
    take_log = scan.axis_params.composite_options.get("take_log", DEFAULT_CMP_TAKE_LOG)
    return theta, subsampling_y, oversampling, take_log


def compute_scan_cor_nabu_growing_window(scan):
    """
    Call to nabu.preproc.alignment.CenterOfRotation from the scan axis_params
    value.

    :param `.TomoBase` scan: scan to process
    :raises: ValueError if half acquisition with an odd number of projections
    :return: Union[float, None]
    """
    if not scan.axis_params.use_sinogram:
        radio_1, radio_2 = AxisProcess.get_inputs(scan=scan)
        if radio_1 is None or radio_2 is None:
            raise NoAxisUrl("Unable to find projections for nabu axis calculation")
    else:
        radio_1, radio_2 = None, None

    _logger.info(
        "compute scan axis from nabu CenterOfRotationGrowingWindow with padding "
        "mode {} and side {}. Use sinogram: {}".format(
            scan.axis_params.padding_mode,
            scan.axis_params.side,
            scan.axis_params.use_sinogram,
        )
    )

    if scan.axis_params.use_sinogram:
        return compute_cor_nabu_growing_window_sinogram(scan=scan)
    else:
        return compute_cor_nabu_growing_window(
            radio_1=radio_1.copy(),
            radio_2=radio_2.copy(),
            side=scan.axis_params.side,
            padding_mode=scan.axis_params.padding_mode,
            flip_frame_2_lr=scan.axis_params.flip_lr,
        )


def compute_cor_nabu_sliding_window(
    radio_1: numpy.ndarray,
    radio_2: numpy.ndarray,
    side: str,
    padding_mode,
    flip_frame_2_lr=True,
    horz_fft_width=False,
):
    """
    Call nabu.preproc.alignement.CenterOfRotationSlidingWindow.find_shift

    :param radio_1:
    :param radio_2:
    :param padding_mode:
    :param str side: side of the cor
    :param horz_fft_width:
    :param bool flip_frame_2_lr: if True will left-right flip the second frame
    :param half_acq_cor_guess: The approximate position of the rotation axis
                               from the image center. Optional. When given a
                               special algorithm is used which can work also
                               in half-tomo conditions.

    :return:
    """
    nabu_class = CenterOfRotationSlidingWindow(horz_fft_width=horz_fft_width)
    return nabu_class.find_shift(
        img_1=radio_1,
        img_2=numpy.fliplr(radio_2) if flip_frame_2_lr else radio_2,
        side=side,
        roi_yxhw=None,
        padding_mode=padding_mode,
        median_filt_shape=None,
    )


def compute_cor_nabu_sliding_window_sinogram(
    scan,
):
    """
    Call nabu.preproc.alignement.CenterOfRotationGrowingWindow.find_shift

    :param TomwerScanBase scan:
    :raises: ValueError if half acquisition with an odd number of projections

    :return:
    """
    check_sinogram_half(scan)
    corfinder = SinoCORFinder(
        adapt_tomwer_scan_to_nabu(scan),
        slice_idx=scan.axis_params.sinogram_line,
        subsampling=scan.axis_params.sinogram_subsampling,
        do_flatfield=True,
        cor_options=scan.axis_params.get_nabu_cor_options(),
        logger=_logger,
    )
    res = corfinder.find_cor(method=AxisMode.sliding_window.value)
    # value return is absolute
    if scan.dim_1 is None:
        _logger.warning("unagle to find image width. Set half width to 1024")
        res = res - 1024
    else:
        res = res - scan.dim_1 // 2
    return res


def compute_scan_cor_nabu_sliding_window(scan):
    """
    Call to nabu.preproc.alignment.CenterOfRotation from the scan axis_params
    value.

    :param `.TomoBase` scan: scan to process
    :raises: ValueError if half acquisition with an odd number of projections
    :return: Union[float, None]
    """
    if not scan.axis_params.use_sinogram:
        radio_1, radio_2 = AxisProcess.get_inputs(scan=scan)
        if radio_1 is None or radio_2 is None:
            raise NoAxisUrl("Unable to find projections for nabu axis calculation")
    else:
        radio_1 = radio_2 = None

    _logger.info(
        "compute scan axis from nabu CenterOfRotationSlidingWindow with padding "
        "mode {} and side {}. Use sinogram: {}".format(
            scan.axis_params.padding_mode,
            scan.axis_params.side,
            scan.axis_params.use_sinogram,
        )
    )

    if scan.axis_params.use_sinogram:
        return compute_cor_nabu_sliding_window_sinogram(scan=scan)
    else:
        return compute_cor_nabu_sliding_window(
            radio_1=radio_1.copy(),
            radio_2=radio_2.copy(),
            side=scan.axis_params.side,
            padding_mode=scan.axis_params.padding_mode,
            flip_frame_2_lr=scan.axis_params.flip_lr,
        )


def compute_cor_nabu_centered(
    radio_1: numpy.ndarray,
    radio_2: numpy.ndarray,
    padding_mode,
    flip_frame_2_lr=True,
    horz_fft_width=False,
):
    """
    Call nabu.preproc.alignement.CenterOfRotation.find_shift

    :param radio_1:
    :param radio_2:
    :param padding_mode:
    :param horz_fft_width:
    :param bool flip_frame_2_lr: if True will left-right flip the second frame
    :param half_acq_cor_guess: The approximate position of the rotation axis
                               from the image center. Optional. When given a
                               special algorithm is used which can work also
                               in half-tomo conditions.

    :return:
    """
    nabu_class = CenterOfRotation(horz_fft_width=horz_fft_width)
    return nabu_class.find_shift(
        img_1=radio_1,
        img_2=numpy.fliplr(radio_2) if flip_frame_2_lr else radio_2,
        roi_yxhw=None,
        padding_mode=padding_mode,
        median_filt_shape=None,
    )


def compute_scan_cor_nabu_centered(scan):
    """
    Call to nabu.preproc.alignment.CenterOfRotation from the scan axis_params
    value.

    :param `.TomoBase` scan: scan to process
    :return: Union[float, None]

    """
    assert scan.axis_params is not None
    radio_1, radio_2 = AxisProcess.get_inputs(scan=scan)
    if radio_1 is None or radio_2 is None:
        raise NoAxisUrl("Unable to find projections for nabu axis calculation")

    _logger.info(
        "compute scan axis from nabu CenterOfRotation with padding "
        "mode %s" % scan.axis_params.padding_mode
    )

    return compute_cor_nabu_centered(
        radio_1=radio_1.copy(),
        radio_2=radio_2.copy(),
        padding_mode=scan.axis_params.padding_mode,
        flip_frame_2_lr=scan.axis_params.flip_lr,
    )


def compute_cor_nabu_global(
    radio_1: numpy.ndarray,
    radio_2: numpy.ndarray,
    padding_mode,
    flip_frame_2_lr=True,
    horz_fft_width=False,
):
    """
    Call nabu.preproc.alignement.CenterOfRotation.find_shift

    :param radio_1:
    :param radio_2:
    :param padding_mode:
    :param horz_fft_width:
    :param bool flip_frame_2_lr: if True will left-right flip the second frame
    :return:
    """
    nabu_class = CenterOfRotationAdaptiveSearch(horz_fft_width=horz_fft_width)
    return nabu_class.find_shift(
        img_1=radio_1,
        img_2=numpy.fliplr(radio_2) if flip_frame_2_lr else radio_2,
        roi_yxhw=None,
        padding_mode=padding_mode,
        median_filt_shape=None,
    )


def compute_scan_cor_nabu_global(scan):
    """
    Call to nabu.preproc.alignment.CenterOfRotation from the scan axis_params
    value.

    :param `.TomoBase` scan: scan to process
    :return: Union[float, None]

    """
    assert scan.axis_params is not None
    radio_1, radio_2 = AxisProcess.get_inputs(scan=scan)
    if radio_1 is None or radio_2 is None:
        raise NoAxisUrl("Unable to find projections for nabu axis calculation")

    _logger.info(
        "compute scan axis from nabu CenterOfRotation with padding "
        "mode %s" % scan.axis_params.padding_mode
    )

    return compute_cor_nabu_global(
        radio_1=radio_1.copy(),
        radio_2=radio_2.copy(),
        padding_mode=scan.axis_params.padding_mode,
        flip_frame_2_lr=scan.axis_params.flip_lr,
    )


def get_stdmax_column(x: numpy.ndarray) -> float:
    """

    :param x:
    :return: column index of the maximal standard deviation
    """
    kernel_size = 5
    length = len(x)
    r = range(length - kernel_size)
    y = numpy.empty(length - kernel_size)
    for i in r:
        s = numpy.std(x[i : i + kernel_size])
        y[i] = s

    return y.argmax()


class NoAxisUrl(Exception):
    pass


class AxisProcess(
    Task, SuperviseProcess, input_names=("data",), output_names=("data",)
):
    """
    Process used to compute the center of rotation of a scan

    :param axis_params: parameters to configure the axis process
    :type: AxisRP
    """

    _CALCULATIONS_METHODS = {
        AxisMode.centered: compute_scan_cor_nabu_centered,
        AxisMode.global_: compute_scan_cor_nabu_global,
        AxisMode.sliding_window: compute_scan_cor_nabu_sliding_window,
        AxisMode.growing_window: compute_scan_cor_nabu_growing_window,
        AxisMode.sino_coarse_to_fine: compute_scan_sino_coarse_to_fine,
        AxisMode.composite_coarse_to_fine: compute_scan_composite_coarse_to_fine,
    }

    def __init__(self, process_id=None, varinfo=None, inputs=None, label=None):
        Task.__init__(self, varinfo=varinfo, inputs=inputs, label=label)
        if "recons_params" in inputs:
            raise KeyError("do not use 'recons_params' but use 'axis_params' instead")
        axis_params = inputs.get("axis_params", None)
        if isinstance(axis_params, dict):
            axis_params = AxisRP.from_dict(axis_params)
        elif not (axis_params is None or isinstance(axis_params, AxisRP)):
            raise TypeError(
                "'axis_params' is expected to be None or an instance of AxisRP or a dict"
            )

        SuperviseProcess.__init__(self, process_id=process_id)
        self._mode_calculation_fct = {}
        """dict with function pointer to call for making the mode calculation.
        Function should have only one 'scan' parameter as input"""

        self._axis_params = axis_params or AxisRP()
        """Axis reconstruction parameters to apply"""
        self._locked = False
        """Boolean used to lock reconstruction parameters edition"""
        self._recons_params_before_lock = None
        """Recons parameters register before locking the position"""

    def set_configuration(self, configuration):
        self.set_recons_params(
            recons_params=AxisRP.from_dict(configuration["_rpSetting"])
        )

    def set_recons_params(self, recons_params):
        assert isinstance(recons_params, AxisRP)
        self._axis_params = recons_params

    def lock_position_value(self, lock=True):
        """
        lock the position currently computed or defined by the user.
        In this case we will lock the axis as defined 'fixed' with the current
        value

        :param bool lock: if true lock the currently existing position value
        """
        self._locked = lock
        if lock:
            self._recons_params_before_lock = self._axis_params.to_dict()
            if self._axis_params not in (AxisMode.manual, AxisMode.read):
                self._axis_params.mode = AxisMode.manual
        else:
            if self._recons_params_before_lock:
                self._axis_params.load_from_dict(
                    self._recons_params_before_lock
                )  # noqa

    def run(self):
        """
        Compute the position value then get ready to the next. And call

        .. note:: this simply call `compute`.
                  But this is needed for the AxisProcessThreaded class
        """
        scan = self.inputs.data
        if scan is None:
            return

        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError("input type {} is not managed".format(scan))

        _logger.info("start axis calculation for %s" % scan.path)
        self._axis_params.frame_width = scan.dim_1
        cor = error = None
        try:
            scan_res = self.compute(scan=scan)
        except Exception as e:
            raise e
            scan_res = None
            error = e
        else:
            if isinstance(scan_res, TomwerScanBase):
                cor = scan_res.axis_params.relative_cor_value
            elif scan_res is None:
                if scan.axis_params.relative_cor_value is not None:
                    cor = scan.axis_params.relative_cor_value
            elif isinstance(scan_res, float):
                cor = scan_res
            else:
                assert isinstance(scan_res, dict)
                b_dict = scan_res
                if TomwerScanBase._DICT_AXIS_KEYS in scan_res:
                    b_dict = scan_res["axis_params"]
                cor = b_dict["POSITION_VALUE"]
        finally:
            if cor != "...":
                self._process_end(scan, cor=cor, error=error)
        self.outputs.data = scan_res

    def reprocess(self, scan):
        """
        Reprocess the scan. For now simply call process

        :param scan:
        """
        return self.process(scan=scan)

    def _process_end(self, scan, cor, error=None):
        assert isinstance(scan, TomwerScanBase)
        if scan.process_file is not None:
            entry = "entry"
            if isinstance(scan, HDF5TomoScan):
                entry = scan.entry
            with scan.acquire_process_file_lock():
                self.register_process(
                    process_file=scan.process_file,
                    entry=entry,
                    results={"center_of_rotation": cor if cor is not None else "-"},
                    configuration=self._axis_params.to_dict(),
                    process_index=scan.pop_process_index(),
                    overwrite=True,
                )

        try:
            extra = {
                logconfig.DOC_TITLE: self._scheme_title,
                logconfig.SCAN_ID: scan.path,
            }
            if error is not None:
                info = " ".join(
                    (
                        "fail to compute axis position for scan",
                        str(scan.path),
                        "reason is ",
                        str(error),
                    )
                )
                _logger.processFailed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.FAILED, details=info
                )
            elif scan.axis_params.relative_cor_value is None:
                info = " ".join(
                    ("fail to compute axis position for scan", str(scan.path))
                )
                _logger.processFailed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.FAILED, details=info
                )
            else:
                info = "axis calculation defined for {}: {} (using {})" "".format(
                    str(scan.path),
                    str(scan.axis_params.relative_cor_value),
                    scan.axis_params.mode.value,
                )
                _logger.processSucceed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.SUCCEED, details=info
                )
        except Exception as e:
            _logger.error(e)

    @staticmethod
    def get_inputs_urls(scan):
        """Make sure we have valid projections to be used for axis calculation

        :param TomwerScanBase scan: scan to check
        :raise: NoAxisUrl if fails to found
        :return: the two projections to be used for axis calculation
        :rtype: tuple of AxisResource
        """
        if (
            scan.axis_params
            and scan.axis_params.axis_url_1
            and scan.axis_params.axis_url_1.url
        ):
            return scan.axis_params.axis_url_1, scan.axis_params.axis_url_2
        else:
            _radio_1, _radio_2 = scan.getRadiosForAxisCalc(
                mode=CorAngleMode.manual_selection
            )
        return _radio_1, _radio_2

    @staticmethod
    def get_inputs(scan):
        assert isinstance(scan, TomwerScanBase)
        radio_1, radio_2 = AxisProcess.get_inputs_urls(scan=scan)
        if radio_1 and radio_2:
            mess = " ".join(
                ("input radios are", radio_1.url.path(), "and", radio_2.url.path())
            )
            _logger.info(mess)
            log_ = scan.axis_params.projection_type is ProjectionType.transmission

            # if necessary normalize data
            if radio_1.normalized_data is None:
                radio_1.normalize_data(scan, log_=log_)
            if radio_2.normalized_data is None:
                radio_2.normalize_data(scan, log_=log_)

            if scan.axis_params.paganin_preproc:
                data_1 = radio_1.normalized_data_paganin
                data_2 = radio_2.normalized_data_paganin
            else:
                data_1 = radio_1.normalized_data
                data_2 = radio_2.normalized_data

            if scan.axis_params.scale_img2_to_img1:
                data_2 = image.scale_img2_to_img1(img_1=data_1, img_2=data_2)
            return data_1, data_2
        else:
            _logger.info("fail to find any inputs")
            return None, None

    def compute(self, scan, wait=True):
        """
        Compute the position value for the scan

        :param TomwerScanBase scan:
        :param bool wait: used for threaded process. True if we want to end the
                          computation before releasing hand.
        :return: scan as a TomoBase
        """
        assert scan is not None
        if isinstance(scan, dict):
            _logger.warning("convert scan from a dict")
            _scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            _scan = scan
        assert isinstance(_scan, TomwerScanBase)
        # if the scan has no tomo reconstruction parameters yet create them
        if _scan.axis_params is None:
            _scan.axis_params = AxisRP()

        # copy axis recons parameters. We skip the axis_url which are specific
        # to the scan
        _scan.axis_params.copy(
            self._axis_params, copy_axis_url=False, copy_flip_lr=False
        )
        assert scan.axis_params is not None
        return self._process_computation(scan=_scan)

    def scan_ready(self, scan):
        _logger.info(scan, "processed")

    def _process_computation(self, scan):
        """

        :param TomwerScanBase scan: scan for which we want to compute the axis
                              position.
        :return: scan as a TomoBase or a dict if _return_dict activated
        """
        _logger.info("compute center of rotation for %s" % scan.path)
        try:
            position = self.compute_axis_position(scan)
        except NotImplementedError as e:
            scan.axis_params.set_relative_value(None)
            raise e
        except ValueError as e:
            scan_name = scan.path or "undef scan"
            scan.axis_params.set_relative_value(None)
            raise Exception(
                "Fail to compute axis position for {} reason is {}"
                "".format(scan_name, e)
            )
        else:
            scan.axis_params.set_relative_value(position)
            self._axis_params.frame_width = scan.dim_1
            self._axis_params.set_relative_value(position)
            scan_name = scan.path or "undef scan"
            if scan.axis_params.use_sinogram:
                method = "sinogram"
            else:
                method = scan.axis_params.mode.value
            mess = (
                "Compute axis position ({value}) with {method} for {scan_name}".format(
                    value=str(scan.axis_params.relative_cor_value),
                    method=method,
                    scan_name=scan_name,
                )
            )
            _logger.info(mess)
        if self._return_dict:
            return scan.to_dict()
        else:
            return scan

    def setMode(self, mode, value):
        if mode is AxisMode.manual:
            self._axis_params.cor_position = value
        else:
            raise NotImplementedError("mode not implemented yet")

    def define_calculation_mode(self, mode, fct_pointer):
        """Register the function to call of the given mode

        :param AxisMode mode: the mode to register
        :param fct_pointer: pointer to the function to call
        """
        self._mode_calculation_fct[mode] = fct_pointer

    def compute_axis_position(self, scan):
        """

        :param scan: scan for which we compute the center of rotation
        :type: TomoScan
        :return: position of the rotation axis. Use the `.AxisMode` defined
                 by the `.ReconsParams` of the `.AxisProcess`
        :rtype: float or None (if fail to compute the axis position)
        """
        mode = self._axis_params.mode
        if mode in (AxisMode.manual, AxisMode.read):
            # If mode is read or manual the position_value is not computed and
            # we will keep the actual one (should have been defined previously)
            res = self._axis_params.relative_cor_value
        elif mode in self._CALCULATIONS_METHODS:
            _logger.info("use radios, mode is %s" % mode.value)
            res = self._CALCULATIONS_METHODS[mode](scan)
        else:
            raise NotImplementedError("Method for", mode, "is not defined")
        return res

    @docstring(Task.program_name)
    @staticmethod
    def program_name():
        return "tomwer_axis"

    @docstring(Task.program_version)
    @staticmethod
    def program_version():
        return tomwer.version.version

    @docstring(Task.definition)
    @staticmethod
    def definition():
        return "Compute center of rotation"

    @staticmethod
    def get_cor_frm_process_file(
        process_file, entry, as_url=False
    ) -> Union[None, float]:
        """
        Read cor position from a tomwer_process file

        :param process_file:
        :param entry:
        :return:
        """
        if entry is None:
            with HDF5File(process_file, "r", swmr=True) as h5f:
                entries = AxisProcess._get_process_nodes(
                    root_node=h5f, process=AxisProcess
                )
                if len(entries) == 0:
                    _logger.info("unable to find a Axis process in %s" % process_file)
                    return None
                elif len(entries) > 1:
                    raise ValueError("several entry found, entry should be " "specify")
                else:
                    entry = list(entries.keys())[0]
                    _logger.info("take %s as default entry" % entry)

        with HDF5File(process_file, "r", swmr=True) as h5f:
            axis_nodes = AxisProcess._get_process_nodes(
                root_node=h5f[entry], process=AxisProcess
            )
            index_to_path = {}
            for key, index in axis_nodes.items():
                index_to_path[index] = key

            if len(axis_nodes) == 0:
                return None
            # take the last processed dark ref
            last_process_index = sorted(list(axis_nodes.values()))[-1]
            last_process_dark = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                _logger.debug(
                    "several processing found for dark-ref,"
                    "take the last one: %s" % last_process_dark
                )

            res = None
            if "results" in h5f[last_process_dark].keys():
                results_node = h5f[last_process_dark]["results"]
                if "center_of_rotation" in results_node.keys():
                    if as_url:
                        res = DataUrl(
                            file_path=process_file,
                            data_path="/".join((results_node, "center_of_rotation")),
                            scheme="h5py",
                        )
                    else:
                        res = h5py_read_dataset(results_node["center_of_rotation"])
            return res



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/04/2019"


import shutil
import tempfile
import unittest
from tomwer.core.utils.scanutils import MockEDF
from tomwer.core.utils import getParametersFromParOrInfo
from ..reconstruction.ftseries import Ftseries
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.process.reconstruction.ftseries.params import ReconsParams
from tomwer.test.utils import UtilsTest
import os


class TestFtseriesIO(unittest.TestCase):
    """Test inputs and outputs types of the handler functions"""

    def setUp(self) -> None:
        self.scan_folder = tempfile.mkdtemp()

        self.scan = MockEDF.mockScan(
            scanID=self.scan_folder, nRadio=10, nRecons=1, nPagRecons=4, dim=10
        )
        self.recons_params = ReconsParams()

    def tearDown(self) -> None:
        shutil.rmtree(self.scan_folder)

    def testInputOutput(self) -> None:
        """Test that io using TomoBase instance work"""
        for input_type in (dict, TomwerScanBase):
            for return_dict in (True, False):
                with self.subTest(
                    return_dict=return_dict,
                    input_type=input_type,
                ):
                    input_obj = self.scan
                    if input_obj is dict:
                        input_obj = input_obj.to_dict()
                    ftseries_process = Ftseries(
                        inputs={
                            "pyhst2_params": self.recons_params,
                            "mock_mode": True,
                            "return_dict": return_dict,
                            "data": input_obj,
                        }
                    )
                    ftseries_process.run()
                    out = ftseries_process.outputs.data
                    if return_dict:
                        self.assertTrue(isinstance(out, dict))
                    else:
                        self.assertTrue(isinstance(out, TomwerScanBase))


class TestFtseriesAxis(unittest.TestCase):
    """Test the behavior of ftseries depending on the axis parameter"""

    def setUp(self) -> None:
        self.outputDir = tempfile.mkdtemp()
        self.dataSetID = "scan_3_"
        self.dataDir = UtilsTest.getEDFDataset(self.dataSetID)
        self.datasetDir = os.path.join(self.outputDir, self.dataSetID)
        shutil.copytree(src=os.path.join(self.dataDir), dst=self.datasetDir)
        self.recons_params = ReconsParams()
        self.parFile = os.path.join(self.datasetDir, self.dataSetID + ".par")
        if os.path.exists(self.parFile):
            os.remove(self.parFile)
        self.scan = EDFTomoScan(self.datasetDir)
        self.axis_frm_tomwer_file = os.path.join(self.datasetDir, "correct.txt")

    def tearDown(self) -> None:
        shutil.rmtree(self.outputDir)

    def testAxisCorrectionOption(self) -> None:
        """Test that the pyhst parameter 'DO_AXIS_CORRECTION' is correctly set"""
        self.recons_params.axis.do_axis_correction = False
        ftseries_process = Ftseries(
            inputs={
                "pyhst2_params": self.recons_params,
                "data": self.scan,
            }
        )

        ftseries_process.run()
        self.assertTrue(os.path.exists(self.parFile))
        par_info = getParametersFromParOrInfo(self.parFile)
        self.assertTrue("DO_AXIS_CORRECTION".lower() in par_info)

        self.recons_params.axis.do_axis_correction = True
        ftseries_process.run()
        self.assertTrue(os.path.exists(self.parFile))
        par_info = getParametersFromParOrInfo(self.parFile)
        self.assertFalse(os.path.exists(self.axis_frm_tomwer_file))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys
from silx.gui import qt
import argparse
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.gui.reconstruction.ftserie import FtserieWidget
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.resource import increase_max_number_file
import signal

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


def getinputinfo():
    return "tomwer ftseries [scan_path]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


def main(argv):
    import os

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path", help="Data file to show (h5 file, edf files, spec files)"
    )
    parser.add_argument(
        "--entry", default=None, help="An entry can be specify for hdf5 files"
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )
    parser.add_argument(
        "--dry-run",
        dest="dry_run",
        action="store_true",
        default=False,
        help="Only create the .par file and print the command executed",
    )
    options = parser.parse_args(argv[1:])

    increase_max_number_file()

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication([])

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler
    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    splash = getMainSplashScreen()
    options.scan_path = options.scan_path.rstrip(os.path.sep)

    scan = ScanFactory.create_scan_object(
        scan_path=options.scan_path, entry=options.entry
    )
    scan.set_process_index_frm_tomwer_process_file()

    widget = FtserieWidget()
    widget.setScan(scan=scan)
    # for the application we run for the reconstruction to be finished
    # to give back hand to the user
    widget.reconstruction_stack._forceSync = True

    widget.setDryRun(options.dry_run)
    widget.setWindowTitle("FtSerie")
    widget.setWindowIcon(icons.getQIcon("tomwer"))
    splash.finish(widget)
    widget.show()
    app.exec_()


if __name__ == "__main__":
    main(sys.argv)



# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/03/2017"

"""Module linking QtObjects with an Octave reconstruction parameter from a
specific H5StructEditor"""


class LinkQObject:
    """Class used to store the QObject associated to an H5 parameter"""

    def __init__(self, qobject):
        self.qobject = qobject


class LinkComboBox(LinkQObject):
    """
    Class used to store the Combobox associated to an H5 parameter

    :param qcombobox: the QCombobox in the interface to edit/display the
        given h5ParamName
    :param h5ParamName: the parameter ID in the H5 structure.
    :param fitwithindex: true if we want to store the index in the h5File,
        otherwise store the string
    :param dic: if not None then instead for storing index or string we will
        store the associated value with this key. Keys are the one displayed in
        the QComboBox, values the one in the H5file
    :param filters: filter to apply on the combobox value to get the requested
        value (only apply on load/getting values)
    :param setDefault: if value to set not found in the existing values then
        set the default value, If False add the value to the QComboBox
    """

    def __init__(
        self, qcombobox, fitwithindex, dic=None, filters=None, setDefault=True
    ):
        LinkQObject.__init__(self, qcombobox)
        assert dic is None or type(dic) is dict
        # filter only apply if we want to get the combobox value
        assert filters is None or fitwithindex is False
        self.fitwithindex = fitwithindex
        self.dicCBtoH5 = dic
        self.dicH5ToCB = dict((v, k) for k, v in iter(dic.items())) if dic else None
        self.filters = filters
        self.setDefault = setDefault


class LinkCheckBox(LinkQObject):
    """
    Class used to store the QCheckBox associated to an H5 parameter

    :param qcheckbox: the QCheckBox in the interface to edit/display the
        given h5ParamName
    :param h5ParamName: the parameter ID in the H5 structure.
    :param invert: true if we will not store the value state == checked but
        state == unchecked
    """

    def __init__(self, qcheckbox, invert):
        LinkQObject.__init__(self, qcheckbox)
        self.invert = invert


class LinkLineEdit(LinkQObject):
    """
    Class used to store the QCheckBox associated to an H5 parameter

    :param qlineedit: the QCheckBox in the interface to edit/display the
        given h5ParamName
    :param h5ParamName: the parameter ID in the H5 structure.
    :param h5paramtype: Define the type of the output value
    """

    def __init__(self, qlineedit, h5paramtype):
        LinkQObject.__init__(self, qlineedit)
        self.h5paramtype = h5paramtype


class LinkSelLineEdit(LinkQObject):
    """
    Class used to store the QCheckBox associated to an H5 parameter

    :param qlineedit: the QCheckBox in the interface to edit/display the
        given h5ParamName
    :param h5ParamName: the parameter ID in the H5 structure.
    :param h5paramtype: Define the type of the output value
    """

    def __init__(self, qlineedit, h5paramtype):
        LinkQObject.__init__(self, qlineedit)
        self.h5paramtype = h5paramtype


class LinkGroup(LinkQObject):
    """
    Class used to store an H5 reconstruction parameter with a simple getter and
    setter to call

    :param qlineedit: the QCheckBox in the interface to edit/display the
        given h5ParamName
    :param h5ParamName: the parameter ID in the H5 structure.
    :param setter: the function to set the variable in the GUI
    :param getter: the function to get the variable value
    """

    def __init__(self, group, getter, setter):
        LinkQObject.__init__(self, group)
        self.getter = getter
        self.setter = setter



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "17/01/2017"

import logging
import os
import shutil
import tempfile
import time
import unittest

from silx.gui import qt

from orangecontrib.tomwer.test.TestAcquisition import Simulation
from orangecontrib.tomwer.widgets.control.DataWatcherOW import DataWatcherOW
from tomwer.core.process.control.datawatcher import status as datawatcherstatus
from tomwer.gui.qtapplicationmanager import QApplicationManager

logging.disable(logging.INFO)

# Makes sure a QApplication exists
_qapp = QApplicationManager()


class DataWatcherWaiter(object):
    """Define a simple objecy able to wait for some state of the DataWatcher
    arrive"""

    def __init__(self):
        self.reset()
        self.lastStatus = []

    def reset(self):
        self.lastStatus = []

    def stateHasChanged(self, val):
        """Register all status"""
        if val not in self.lastStatus:
            self.lastStatus.append(val)

    def waitForState(self, state, maxWaiting):
        """simple function wich wait until the DataWatcherWidget reach the given
        state.
        If the widget doesn't reach this state after maxWaiting second. Then fail.

        :param str state: the state we are waiting for
        :param int maxWaiting: the maximal number of second to wait until failling.
        """
        while state not in self.lastStatus:
            time.sleep(1.0)
            _qapp.processEvents()
            maxWaiting -= 1
            if maxWaiting <= 0:
                return False
        return state in self.lastStatus


class TestDataWatcherAcquisition(unittest.TestCase, DataWatcherWaiter):
    """Functions testing the classical behavior of data watcher
    - signal acquisition is over only when all files are copied
    """

    def tearDown(self):
        while _qapp.hasPendingEvents():
            _qapp.processEvents()
        self.dataWatcherWidget.setObservation(False)
        self.dataWatcherWidget.close()
        del self.dataWatcherWidget
        self.dataWatcherWidget = None
        self.s.wait()
        del self.s
        super(TestDataWatcherAcquisition, self).tearDown()
        if os.path.isdir(self.inputdir):
            shutil.rmtree(self.inputdir)

    def setUp(self):
        self.manipulationId = "test10"

        super(TestDataWatcherAcquisition, self).setUp()
        self.inputdir = tempfile.mkdtemp()
        DataWatcherWaiter.reset(self)
        self.dataWatcherWidget = DataWatcherOW(displayAdvancement=False)
        self.dataWatcherWidget.srcPattern = ""
        self.dataWatcherWidget.sigTMStatusChanged.connect(self.stateHasChanged)
        self.dataWatcherWidget.setAttribute(qt.Qt.WA_DeleteOnClose)

        self.s = Simulation(
            self.inputdir,
            self.manipulationId,
            finalState=Simulation.advancement["acquisitionRunning"],
        )

    def testStartAcquisition(self):
        """Make sure the data watch detect the acquisition of started"""
        observeDir = os.path.join(self.inputdir, self.manipulationId)
        for folder in (self.inputdir, observeDir):
            if not os.path.exists(folder):
                os.makedirs(folder)

        self.assertTrue(os.path.isdir(observeDir))

        self.s.createFinalXML(True)
        self.dataWatcherWidget.setFolderObserved(observeDir)

        self.assertTrue(self.dataWatcherWidget.currentStatus == "not processing")
        self.s.start()
        self.s.advanceTo("acquisitionDone")
        self.s.start()
        self.s.wait()
        self.dataWatcherWidget.startObservation()
        self.dataWatcherWidget._widget.observationThread.wait()
        self.dataWatcherWidget._widget.observationThread.observations.dict[
            observeDir
        ].wait()
        finishedAcqui = (
            self.dataWatcherWidget._widget.observationThread.observations.ignoredFolders
        )
        _qapp.processEvents()
        self.assertTrue(observeDir in finishedAcqui)


class TestDataWatcherInteraction(unittest.TestCase):
    """Simple unit test to test the start/stop observation button action"""

    def setUp(self):
        super(TestDataWatcherInteraction, self).setUp()
        self.inputdir = tempfile.mkdtemp()
        self.dataWatcherWidget = DataWatcherOW(displayAdvancement=False)
        self.dataWatcherWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.dataWatcherWidget.srcPattern = ""

    def tearDown(self):
        while _qapp.hasPendingEvents():
            _qapp.processEvents()
        self.dataWatcherWidget.close()
        del self.dataWatcherWidget
        super(TestDataWatcherInteraction, self).tearDown()
        if os.path.isdir(self.inputdir):
            shutil.rmtree(self.inputdir)

        def tearDown(self):
            while _qapp.hasPendingEvents():
                _qapp.processEvents()

    def testStartAndStopAcquisition(self):
        """test multiple start and stop action on the start observation to
        make sure no crash are appearing
        """
        try:
            self.dataWatcherWidget.setFolderObserved(self.inputdir)
            self.dataWatcherWidget.show()
            self.dataWatcherWidget.setObservation(True)
            for iTest in range(5):

                def tearDown(self):
                    while _qapp.hasPendingEvents():
                        _qapp.processEvents()

                self.dataWatcherWidget._widget._qpbstartstop.pressed.emit()
            self.assertTrue(True)
        except Exception:
            self.assertTrue(False)


class WaitForXMLOption(unittest.TestCase, DataWatcherWaiter):
    """test the behavior of datawatcher when the option 'wait for xml copy'
    Basically in this case DataWatcherDirObserver will wait until an .xml
    arrived
    """

    @classmethod
    def setUpClass(cls):
        cls.dataWatcherWidget = DataWatcherOW(displayAdvancement=False)
        cls.dataWatcherWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        cls.dataWatcherWidget.obsMethod = datawatcherstatus.DET_END_XML
        cls.dataWatcherWidget.srcPattern = ""
        cls.manipulationId = "test10"
        super(WaitForXMLOption, cls).setUpClass()

    def setUp(self):
        self.inputdir = tempfile.mkdtemp()
        self.reset()
        self.dataWatcherWidget.setObservation(False)
        self.dataWatcherWidget.resetStatus()
        super(WaitForXMLOption, self).setUp()

    def tearDown(self):
        while _qapp.hasPendingEvents():
            _qapp.processEvents()
        super(WaitForXMLOption, self).tearDown()
        if os.path.isdir(self.inputdir):
            shutil.rmtree(self.inputdir)

    @classmethod
    def tearDownClass(cls):
        cls.dataWatcherWidget.close()
        del cls.dataWatcherWidget
        if hasattr(cls, "s"):
            cls.s.quit()
            del cls.s
        super(WaitForXMLOption, cls).tearDownClass()

    def testAcquistionNotEnding(self):
        """Check behavior if an acquisition never end"""
        observeDir = os.path.join(self.inputdir, self.manipulationId)
        for folder in (self.inputdir, observeDir):
            if not os.path.exists(folder):
                os.makedirs(folder)

        self.assertTrue(os.path.isdir(observeDir))

        self.s = Simulation(
            self.inputdir,
            self.manipulationId,
            finalState=Simulation.advancement["acquisitionRunning"],
        )
        self.dataWatcherWidget.setFolderObserved(observeDir)
        self.dataWatcherWidget.show()
        self.dataWatcherWidget.sigTMStatusChanged.connect(self.stateHasChanged)

        self.assertTrue(self.dataWatcherWidget.currentStatus == "not processing")
        self.s.start()
        self.s.wait()
        self.dataWatcherWidget.setObservation(True)
        self.dataWatcherWidget._widget.observationThread.wait()
        self.dataWatcherWidget._widget.observationThread.observations.dict[
            observeDir
        ].wait()
        finishedAcqui = (
            self.dataWatcherWidget._widget.observationThread.observations.ignoredFolders
        )
        _qapp.processEvents()
        self.assertFalse(observeDir in finishedAcqui)

    def testAcquistionEnded(self):
        """Check behavior if an acquisition is ending"""
        manipulationId = "test10"
        observeDir = os.path.join(self.inputdir, self.manipulationId)
        for folder in (self.inputdir, observeDir):
            if not os.path.exists(folder):
                os.makedirs(folder)

        self.assertTrue(os.path.isdir(observeDir))

        self.s = Simulation(
            self.inputdir,
            manipulationId,
            finalState=Simulation.advancement["acquisitionDone"],
        )
        self.s.createFinalXML(True)
        self.dataWatcherWidget.setFolderObserved(observeDir)
        self.dataWatcherWidget.show()
        self.dataWatcherWidget.sigTMStatusChanged.connect(self.stateHasChanged)

        self.assertTrue(self.dataWatcherWidget.currentStatus == "not processing")
        self.s.start()
        self.s.wait()
        self.dataWatcherWidget.setObservation(True)
        self.dataWatcherWidget._widget.observationThread.wait()
        self.dataWatcherWidget._widget.observationThread.observations.dict[
            observeDir
        ].wait()
        finishedAcqui = (
            self.dataWatcherWidget._widget.observationThread.observations.ignoredFolders
        )
        _qapp.processEvents()
        self.assertTrue(observeDir in finishedAcqui)


class TestRSync(unittest.TestCase, DataWatcherWaiter):
    """test that the synchronization using RSyncManager is working"""

    def setUp(self):
        super(TestRSync, self).setUp()
        self.inputdir = tempfile.mkdtemp()
        self.outputdir = tempfile.mkdtemp()
        DataWatcherWaiter.reset(self)
        self.dataWatcherWidget = DataWatcherOW(displayAdvancement=False)
        self.dataWatcherWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.dataWatcherWidget._widget.setSrcAndDestPattern(
            self.inputdir, self.outputdir
        )

    def tearDown(self):
        while _qapp.hasPendingEvents():
            _qapp.processEvents()
        self.dataWatcherWidget.close()
        del self.dataWatcherWidget
        if hasattr(self, "s"):
            self.s.quit()
            del self.s
        super(TestRSync, self).tearDown()
        for d in (self.inputdir, self.outputdir):
            if os.path.isdir(d):
                shutil.rmtree(d, ignore_errors=True)

    def testStartAcquisition(self):
        """Test that rsync is launched when an acquisition is discovered"""
        manipulationId = "test10"
        observeDir = os.path.join(self.inputdir, manipulationId)
        for folder in (self.inputdir, observeDir):
            if not os.path.exists(folder):
                os.makedirs(folder)

        self.assertTrue(os.path.isdir(observeDir))

        self.s = Simulation(
            self.inputdir,
            manipulationId,
            finalState=Simulation.advancement["acquisitionRunning"],
        )

        self.s.setSrcDestPatterns(self.inputdir, self.outputdir)
        self.s.createFinalXML(True)
        self.dataWatcherWidget.setFolderObserved(self.inputdir)
        self.dataWatcherWidget.show()
        self.dataWatcherWidget.sigTMStatusChanged.connect(self.stateHasChanged)
        self.assertTrue(self.dataWatcherWidget.currentStatus == "not processing")
        self.dataWatcherWidget.setObservation(True)
        self.s.start()
        # check state scanning
        time.sleep(0.5)

        self.dataWatcherWidget.stopObservation()

        # in this case the .info should be in the output dir also
        test10_output = os.path.join(self.outputdir, "test10")
        test10_input = os.path.join(self.inputdir, "test10")
        self.assertTrue(os.path.isfile(os.path.join(test10_output, "test10.info")))

        # make sure file transfert have been started (using rsync)
        # all file in outputdir should be in input dir
        time.sleep(2)
        # check that some .edf file have already been copied
        self.assertTrue(len(test10_output) > 5)

        # xml shouldn't be there because we are righting it at the end
        self.assertFalse(os.path.isfile(os.path.join(test10_output, "test10.xml")))
        self.assertFalse(os.path.isfile(os.path.join(test10_input, "test10.xml")))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/03/2019"

import tomwer.core.process.reconstruction.axis
from ..utils import WidgetLongProcessing
from orangewidget import gui
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from orangewidget.settings import Setting
from orangecontrib.tomwer.orange.settings import CallbackSettingsHandler
from orangewidget.widget import Input, Output
from processview.gui.processmanager import ProcessManager
from processview.core.manager import DatasetState
from processview.core import helpers as pv_helpers
from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.synctools.stacks.reconstruction.axis import AxisProcessStack
from tomwer.core.process.reconstruction.axis import AxisProcess
from tomwer.synctools.axis import QAxisRP
from tomwer.gui.reconstruction.axis import AxisWindow
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.scan.scanbase import _TomwerBaseDock
from silx.gui import qt
from tomwer.core import settings
from tomwer.core import utils
import functools
import copy

import logging

logger = logging.getLogger(__name__)


class AxisOW(SuperviseOW, WidgetLongProcessing):
    """
    Widget used to defined the center of rotation axis to be used for a
    reconstruction.

    :param bool _connect_handler: True if we want to store the modifications
                      on the setting. Need for unit test since
                      keep alive qt widgets.
    """

    name = "default center of rotation"
    id = "orange.widgets.tomwer.axis"
    description = "use to compute the center of rotation"
    icon = "icons/axis.png"
    priority = 14
    keywords = [
        "tomography",
        "axis",
        "tomwer",
        "reconstruction",
        "rotation",
        "position",
        "center of position",
    ]

    ewokstaskclass = tomwer.core.process.reconstruction.axis.AxisProcess

    want_main_area = True
    resizing_enabled = True
    allows_cycle = True
    compress_signal = False

    settingsHandler = CallbackSettingsHandler()

    sigScanReady = qt.Signal(TomwerScanBase)
    """Signal emitted when a scan is ready"""

    _rpSetting = Setting(dict())
    "AxisRP store as dict"
    # this one is keep for backward compatibility. Will be removed on 0.10 I guess

    static_input = Setting({"data": None, "axis_params": None})

    class Inputs:
        data = Input(name="data", type=TomwerScanBase, doc="one scan to be process")
        data_recompute_axis = Input(
            name="change recons params",
            type=_TomwerBaseDock,
            doc="recompute delta / beta",
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

    def __init__(self, parent=None, axis_params=None):
        """

        :param parent: QWidget parent or None
        :param _connect_handler: used for CI, because if connected fails CI
        :param QAxisRP axis_params: reconstruction parameters
        """
        if axis_params is not None:
            if not isinstance(axis_params, QAxisRP):
                raise TypeError(
                    "axis_params should be an instance of "
                    "QAxisRP. Not {}".format(type(axis_params))
                )
        self._axis_params = axis_params or QAxisRP()

        # handle settings
        axis_params_settings = self.static_input.get("axis_params", None)
        if axis_params_settings is None:
            axis_params_settings = self._rpSetting
        if axis_params_settings != dict():
            try:
                self._axis_params.load_from_dict(axis_params_settings)
            except Exception as e:
                logger.error("fail to load reconstruction settings:", str(e))
        gui_settings = self.static_input.get("gui", None)
        if gui_settings is None:
            gui_settings = self._rpSetting.get("gui", {})

        self.__lastAxisProcessParamsCache = None
        # used to memorize the last (reconstruction parameters, scan)
        self.__scan = None
        self.__skip_exec = False
        self._n_skip = 0
        self._patches = []
        """patches for processing"""

        WidgetLongProcessing.__init__(self)
        SuperviseOW.__init__(self, parent)
        self._processingStack = AxisProcessStack(
            axis_params=self._axis_params, process_id=self.process_id
        )

        self._widget = AxisWindow(parent=self, axis_params=self._axis_params)

        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.setContentsMargins(0, 0, 0, 0)

        self._layout.addWidget(self._widget)

        # expose API
        self.setMode = self._widget.setMode
        self.getRadioMode = self._widget.getRadioMode
        self.getEstimatedCor = self._widget.getEstimatedCor
        self.getAxis = self._widget.getAxis
        self.setValueLock = self._widget.setValueLock
        self._setModeLockFrmSettings = self._widget._setModeLockFrmSettings
        self._setValueLockFrmSettings = self._widget._setValueLockFrmSettings

        # load settings
        if "mode_is_lock" in gui_settings:
            mode_lock = gui_settings["mode_is_lock"]
            # if the mode is manual or read ignore lock
            if not (
                mode_lock is True
                and self.getRadioMode() in (AxisMode.manual, AxisMode.read)
            ):
                self._setModeLockFrmSettings(mode_lock)

        if "value_is_lock" in gui_settings:
            if gui_settings["value_is_lock"] is True:
                self._setValueLockFrmSettings(bool(gui_settings["value_is_lock"]))
        # expose API
        self._applyBut = self._widget._axisWidget._controlWidget._applyBut
        # connect Signal / Slot
        self._widget.sigComputationRequested.connect(self.__compute)
        self._widget.sigApply.connect(self.__validate)
        self._widget.sigAxisEditionLocked.connect(self.__lockReconsParams)
        self._processingStack.sigComputationStarted.connect(self._processingStart)
        self._processingStack.sigComputationEnded.connect(self._scanProcessed)

        self._axis_params.sigChanged.connect(self._updateSettingsVals)
        self._widget.sigAxisEditionLocked.connect(self._updateSettingsVals)
        self._widget.sigMethodChanged.connect(self._updateSettingsVals)
        self._widget.sigLockModeChanged.connect(self._updateSettingsVals)

    def _processingStart(self, *args, **kwargs):
        WidgetLongProcessing._startProcessing(self)

    def _scanProcessed(self, scan, future_scan):
        assert isinstance(scan, TomwerScanBase)
        WidgetLongProcessing._endProcessing(self, scan)
        if self.isValueLock() or self.isModeLocked():
            self.__scan = scan
            self.__validate()
        else:
            pm = ProcessManager()
            # handle previous scan if not validated yey: skip it
            if (
                self.__scan is not None
                and pm.get_dataset_state(dataset=self.__scan, process=self)
                != DatasetState.SUCCEED
            ):
                pm.notify_dataset_state(
                    dataset=self.__scan,
                    process=self,
                    state=DatasetState.SKIPPED,
                    details="Scan was waiting for validation. Has been replaced by another scan. No scan stack on Axis process",
                )

            # handle the new scan
            # retrieve details to keep them in `memory`
            self.__scan = scan
            details = pm.get_dataset_details(dataset=self.__scan, process=self)
            details = "Wait for user validation. " + details
            pm.notify_dataset_state(
                dataset=self.__scan,
                process=self,
                state=DatasetState.WAIT_USER_VALIDATION,
                details=details,
            )
            self.activateWindow()
            self.raise_()
            self.show()

    def __compute(self):
        if self.__scan:
            dict_for_cache = self._axis_params.to_dict().copy()
            dict_for_cache.pop("POSITION_VALUE")
            params_cache = (
                dict_for_cache,
                str(self.__scan),
            )
            # check mode is not locked or manual and value is still here
            if (
                params_cache == self.__lastAxisProcessParamsCache
            ) and self._processingStack.is_computing():
                logger.error(
                    "Parameters and scan are the same as last request. Please wait until the processing is done."
                )
            else:
                self.__lastAxisProcessParamsCache = params_cache
                callback = functools.partial(self._updatePosition, self.__scan)
                self._processingStack.add(
                    scan=self.__scan,
                    configuration=self._axis_params.to_dict(),
                    callback=callback,
                )

    def __validate(self):
        """Validate the current scan and move the scan to the next process.
        The Axis will process the next scan in the stack.
        """
        if self.__scan:
            if self.getRadioMode() is AxisMode.manual:
                if (
                    self.__scan._axis_params.frame_width is None
                    and self.__scan.dim_1 is not None
                ):
                    self.__scan._axis_params.frame_width = self.__scan.dim_1

                self.__scan._axis_params.set_relative_value(
                    self._axis_params.relative_cor_value
                )
                pv_helpers.notify_succeed(
                    process=self,
                    dataset=self.__scan,
                    details="axis calculation defined for {}: {} (using {})"
                    "".format(
                        str(self.__scan.path),
                        str(self.__scan.axis_params.relative_cor_value),
                        "manual",
                    ),
                )
            # validate the center of rotation
            pm = ProcessManager()
            # retrieve details to keep them in memory
            details = pm.get_dataset_details(dataset=self.__scan, process=self)
            state = pm.get_dataset_state(dataset=self.__scan, process=self)
            if state in (DatasetState.ON_GOING, DatasetState.WAIT_USER_VALIDATION):
                # update the state to SUCCEED
                ProcessManager().notify_dataset_state(
                    dataset=self.__scan,
                    process=self,
                    state=DatasetState.SUCCEED,
                    details=details,
                )
            self.accept()
            self.scan_ready(scan=self.__scan)
        self.hide()

    def __lockReconsParams(self, lock):
        self.lock_position_value(lock)

    def scan_ready(self, scan):
        assert isinstance(scan, TomwerScanBase)
        self.Outputs.data.send(scan)
        self.sigScanReady.emit(scan)

    def _informNoProjFound(self, scan):
        msg = qt.QMessageBox(self)
        msg.setIcon(qt.QMessageBox.Warning)
        text = (
            "Unable to find url to compute the axis of `%s`" % scan.path
            or "no path given"
        )
        text += ", please select them from the `axis input` tab"
        msg.setText(text)
        msg.exec_()

    def _updateSettingsVals(self):
        self._rpSetting = self._axis_params.to_dict()
        self._rpSetting["gui"] = {
            "mode_is_lock": self.isModeLocked(),
            "value_is_lock": self.isValueLock(),
        }
        self.static_input = {
            "data": None,
            "axis_params": self._axis_params.to_dict(),
            "gui": {
                "mode_is_lock": self.isModeLocked(),
                "value_is_lock": self.isValueLock(),
            },
        }

    def _skip_exec(self, b):
        """util function used for unit test. If activate, skip the call to
        self.exec() in process"""
        self.__skip_exec = b

    @property
    def recons_params(self):
        return self._axis_params

    def _instanciateReconsParams(self):
        return QAxisRP()

    def _lock_axis_controls(self, lock):
        """

        :param bool lock: lock the axis controls to avoid modification of the
                          requested options, method... of the axis calculation
                          when this value is under calculation.
        """
        self._widget.setLocked(lock)

    def isValueLock(self):
        """
        Check if the cor value has been lock. If so we simply copy the cor
        value and move to the next scan
        """
        return self._widget.isValueLock()

    def isModeLocked(self):
        """
        Check if the mode has been lock or not. If lock then call the
        algorithm and does not wait for any user feedback
        """
        return self._widget.isModeLock()

    @Inputs.data
    def new_data_in(self, scan):
        if scan is None:
            return
        scan_ = copy.copy(scan)
        if not (
            settings.isOnLbsram(scan)
            and utils.isLowOnMemory(settings.get_lbsram_path())
        ):
            set_position = False  # avoid shift reset
            self._widget.setScan(scan=scan_, set_position=set_position)
        elif scan_.axis_params is None:
            scan_.axis_params = QAxisRP()
        self.process(scan=scan_)

    def process(self, scan):
        if scan is None:
            return
        self.__scan = scan
        self._axis_params.frame_width = scan.dim_1
        if (
            settings.isOnLbsram(scan)
            and utils.isLowOnMemory(settings.get_lbsram_path()) is True
        ):
            self._updatePosition(scan=scan)
            self.scan_ready(scan=scan)
        elif self.__skip_exec:
            self._n_skip += 1
            if self.isValueLock():
                scan._axis_params.set_relative_value(
                    self._axis_params.relative_cor_value
                )
                cor = scan._axis_params.relative_cor_value
                if isinstance(scan, HDF5TomoScan):
                    entry = scan.entry
                else:
                    entry = "entry"
                with scan.acquire_process_file_lock():
                    ap = AxisProcess(
                        process_id=self.process_id,
                        inputs={
                            "data": None,
                        },
                    )
                    try:
                        AxisProcess._register_process(
                            process_file=scan.process_file,
                            entry=entry,
                            process=ap,
                            results={
                                "center_of_rotation": cor if cor is not None else "-"
                            },
                            configuration=self._axis_params.to_dict(),
                            process_index=scan.pop_process_index(),
                            overwrite=True,
                        )
                    except Exception as e:
                        logger.warning(f"Fail to register Axis process. Error is {e}")
                    ProcessManager().notify_dataset_state(
                        dataset=scan,
                        process=ap,
                        state=DatasetState.SUCCEED,
                    )
            else:
                processing_class = AxisProcess(
                    inputs={
                        "axis_params": self._axis_params,
                        "data": self.__scan,
                    }
                )
                for patch in self._patches:
                    mode, function = patch
                    processing_class._CALCULATIONS_METHODS[mode] = function
                processing_class.run()
            self._updatePosition(scan=self.__scan)
            if self.isModeLocked() or self.isValueLock():
                self.scan_ready(scan=scan)

        elif self.isValueLock():
            entry = scan.entry if isinstance(scan, HDF5TomoScan) else "entry"
            cor = self._axis_params.relative_cor_value
            scan.axis_params.set_relative_value(cor)
            scan._axis_params.mode = "manual"
            try:
                AxisProcess._register_process(
                    process_file=scan.process_file,
                    entry=entry,
                    process=AxisProcess,
                    results={"center_of_rotation": cor if cor is not None else "-"},
                    configuration=self._axis_params.to_dict(),
                    process_index=scan.pop_process_index(),
                    overwrite=True,
                )
            except Exception as e:
                logger.warning(f"Fail to register Axis process. Error is {e}")
            ProcessManager().notify_dataset_state(
                dataset=scan,
                process=AxisProcess(
                    process_id=self.process_id,
                    inputs={
                        "data": None,
                    },
                ),
                state=DatasetState.SUCCEED,
            )
            self.scan_ready(scan=scan)

        elif self.isModeLocked():
            callback = functools.partial(self._updatePosition, scan)
            self._processingStack.add(
                scan=scan, configuration=self._axis_params.to_dict(), callback=callback
            )
        else:
            self.activateWindow()
            self.raise_()
            self.show()

    def _updatePosition(self, scan):
        if scan.axis_params is not None:
            self._widget.setPosition(
                frm=None, value=scan.axis_params.relative_cor_value
            )

    @Inputs.data_recompute_axis
    def reprocess(self, scan):
        """Recompute the axis for scan"""
        if scan is not None:
            # for now The behavior for reprocessing is the sama as for processing
            if hasattr(scan, "instance"):
                self.process(scan=scan.instance)
            else:
                self.process(scan=scan)

    def close(self):
        self._processingStack.stop()
        self._processingStack.wait_computation_finished()
        self._widget = None
        self._processingStack = None
        super().close()

    def patch_calc_method(self, mode, function):
        self._patches.append((mode, function))

    def keyPressEvent(self, event):
        """The event has to be filtered since we have some children
         that can be edited using the 'enter' key as defining the cor manually
        (see #481)). As we are in a dialog this automatically trigger
        'accepted'. See https://forum.qt.io/topic/5080/preventing-enter-key-from-triggering-ok-in-qbuttonbox-in-particular-qlineedit-qbuttonbox/5
        """
        if event.key() != qt.Qt.Key_Enter:
            super().keyPressEvent(event)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "31/03/2021"


from silx.gui import qt
from tomwer.core.scan.scanbase import TomwerScanBase
import weakref
import logging
from tomwer.core.scan.edfscan import EDFTomoScan

_logger = logging.getLogger(__name__)


class ScanOverviewWidget(qt.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._scan = None
        self.setLayout(qt.QVBoxLayout())
        self._tree = qt.QTreeWidget(self)
        self._tree.setColumnCount(2)
        self._tree.setHeaderLabels(("entry", "value"))
        self.layout().addWidget(self._tree)

        # 1: define instrument
        self._instrument = qt.QTreeWidgetItem(self._tree)
        self._instrument.setText(0, "instrument")
        # 1.1 define beam
        self._beam = qt.QTreeWidgetItem(self._instrument)
        self._beam.setText(0, "beam")
        self._energy = qt.QTreeWidgetItem(self._beam)
        self._energy.setText(0, "energy")
        # 1.1 define detector
        self._frames = qt.QTreeWidgetItem(self._instrument)
        self._frames.setText(0, "frames")
        self._projections = qt.QTreeWidgetItem(self._frames)
        self._projections.setText(0, "projections")
        self._darks = qt.QTreeWidgetItem(self._frames)
        self._darks.setText(0, "darks")
        self._flats = qt.QTreeWidgetItem(self._frames)
        self._flats.setText(0, "flats")
        self._alignments = qt.QTreeWidgetItem(self._frames)
        self._alignments.setText(0, "alignments")

        self._x_pixel_size = qt.QTreeWidgetItem(self._instrument)
        self._x_pixel_size.setText(0, "x pixel size")
        self._y_pixel_size = qt.QTreeWidgetItem(self._instrument)
        self._y_pixel_size.setText(0, "y pixel size")

        # 2: define sample
        self._sample = qt.QTreeWidgetItem(self._tree)
        self._sample.setText(0, "sample")
        self._sample_name = qt.QTreeWidgetItem(self._sample)
        self._sample_name.setText(0, "name")

        # 3: other hight level items
        self._startTime = qt.QTreeWidgetItem(self._tree)
        self._startTime.setText(0, "start_time")
        self._endTime = qt.QTreeWidgetItem(self._tree)
        self._endTime.setText(0, "end_time")
        self._title = qt.QTreeWidgetItem(self._tree)
        self._title.setText(0, "title")

        # set up
        self._instrument.setExpanded(True)
        self._frames.setExpanded(True)
        self._sample.setExpanded(True)
        self._beam.setExpanded(True)

    def setScan(self, scan):
        if scan is None:
            self._scan = scan
            self.clear()
        elif not isinstance(scan, TomwerScanBase):
            raise TypeError(f"{scan} is expected to be an instance of {TomwerScanBase}")
        else:
            self._scan = weakref.ref(scan)
        self.update_tree()

    def getScan(self):
        if self._scan is None or self._scan() is None:
            return None
        else:
            return self._scan()

    def update_tree(self):
        if self.getScan() is not None:
            for fct in (
                self._updateInstrument,
                self._updateSample,
                self._updateTimes,
                self._updateNames,
            ):
                try:
                    fct()
                except Exception as e:
                    _logger.info(e)
            self._tree.resizeColumnToContents(0)

    def _updateInstrument(self):
        self._updateFrames()
        self._updateEnergy()
        self._updatePixelSize()

    def _setColoredTxt(
        self, item, text, column=1, hightlight_red=False, hightlight_orange=False
    ):
        if text in (None, str(None)):
            text = "?"
        if hightlight_red:
            bkg_color = qt.QColor(220, 0, 0, 200)
        elif hightlight_orange:
            bkg_color = qt.QColor(200, 160, 0, 150)
        else:
            bkg_color = qt.QColor(0, 220, 0, 50)

        item.setText(column, text)
        item.setBackground(0, qt.QBrush(bkg_color))

    def _updateSample(self):
        pass

    def _updateTimes(self):
        scan = self.getScan()
        self._startTime.setText(1, str(scan.start_time))
        self._endTime.setText(1, str(scan.end_time))

    def _updateFrames(self):
        scan = self.getScan()
        assert isinstance(scan, TomwerScanBase)
        # frames
        n_frames = len(scan.frames)
        self._setColoredTxt(
            item=self._frames,
            text=str(n_frames),
            hightlight_red=(n_frames in (0, None)),
        )
        # projections
        n_proj = len(scan.projections)
        self._setColoredTxt(
            item=self._projections,
            text=str(n_proj),
            hightlight_red=(n_proj in (0, None)),
        )
        # darks
        n_darks = len(scan.darks)
        self._setColoredTxt(
            item=self._darks,
            text=str(n_darks),
            hightlight_red=(n_darks in (0, None)),
        )

        # flats
        n_flats = len(scan.flats)
        self._setColoredTxt(
            item=self._flats,
            text=str(n_flats),
            hightlight_red=(n_flats in (0, None)),
        )
        # align
        n_alignment = len(scan.alignment_projections)
        self._setColoredTxt(
            item=self._alignments,
            text=str(n_alignment),
        )

    def _updateEnergy(self):
        scan = self.getScan()
        assert isinstance(scan, TomwerScanBase)
        energy = scan.energy
        self._setColoredTxt(
            item=self._energy,
            text=f"{energy} (kev)",
            hightlight_red=energy in (0, None),
        )

    def _updateNames(self):
        scan = self.getScan()
        assert isinstance(scan, TomwerScanBase)
        sample_name = scan.sample_name
        sequence_name = scan.sequence_name
        self._title.setText(1, sequence_name)
        self._sample_name.setText(1, sample_name)

    def _updateTomoN(self):
        scan = self.getScan()
        assert isinstance(scan, TomwerScanBase)
        tomo_n = scan.tomo_n
        self._setColoredTxt(
            item=self._tomoNQLE,
            text=str(tomo_n),
        )

    def _updateScanRange(self):
        scan = self.getScan()
        assert isinstance(scan, TomwerScanBase)
        scan_range = scan.scan_range
        self._setColoredTxt(
            item=self._scanRangeQLE,
            text=str(scan_range),
        )

    def _updatePixelSize(self):
        scan = self.getScan()
        assert isinstance(scan, TomwerScanBase)
        if isinstance(scan, EDFTomoScan):
            x_pixel_size = y_pixel_size = scan.pixel_size
        else:
            x_pixel_size = scan.x_pixel_size
            y_pixel_size = scan.y_pixel_size
        self._setColoredTxt(
            item=self._x_pixel_size,
            text=f"{x_pixel_size} (m)",
            hightlight_red=x_pixel_size in (None, 0.0, 1.0),
        )
        self._setColoredTxt(
            item=self._y_pixel_size,
            text=f"{y_pixel_size} (m)",
            hightlight_red=y_pixel_size in (None, 0.0, 1.0),
        )

    def clear(self):
        self._tree.clear()



# coding: utf-8
###########################################################################
# Copyright (C) 2016-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "16/11/2020"


from silx.gui import qt
from tomwer.core.process.control.datalistener import DataListener
from ..processingstack import FIFO, ProcessingThread
from tomwer.core.scan.blissscan import BlissScan

try:
    # new HDF5Config class
    from nxtomomill.io.config import TomoHDF5Config as HDF5Config
except ImportError:
    from nxtomomill.io.config import HDF5Config
import logging

_logger = logging.getLogger(__name__)


class DataListenerProcessStack(FIFO, qt.QObject):
    """Stack of file conversion once received by the data-listener
    from a bliss file and a specific entry"""

    def __init__(self, parent=None):
        qt.QObject.__init__(self, parent=parent)
        FIFO.__init__(self, process_id=-1)
        self._results = {}

    def _process(self, scan, configuration, callback=None):
        assert isinstance(scan, BlissScan)
        assert isinstance(configuration, HDF5Config)
        _logger.info("DataListenerProcessStack is processing {}".format(str(scan)))
        self._scan_currently_computed = scan
        self._computationThread.finished.connect(self._end_threaded_computation)

        self._computationThread.init(scan=scan, configuration=configuration)
        # need to manage connect before starting it because
        self._computationThread.start()

    def _end_threaded_computation(self, callback=None):
        self._computationThread.finished.disconnect(self._end_threaded_computation)
        super()._end_threaded_computation(callback=callback)

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        thread = _DataListenerConverterThread()
        thread.setParent(self)
        return thread

    def _end_computation(self, scan, future_scan, callback):
        """
        callback when the computation thread is finished

        :param scan: pass if no call to '_computationThread is made'
        """
        if callback is not None:
            callback()
        if scan in self._results:
            nx_scan = self._results[scan]
            del self._results[scan]
        else:
            nx_scan = None
        self.sigComputationEnded.emit(nx_scan, None)
        self._processing = False
        if self.can_process_next():
            self._process_next()

    def register_result(self, bliss_scan, nx_scan):
        self._results[bliss_scan] = nx_scan


class _DataListenerConverterThread(ProcessingThread):
    """
    Thread use to execute the processing of nxtomomill
    """

    def __init__(self):
        ProcessingThread.__init__(self)
        self._scan = None
        self._configuration = None

    def init(self, scan, configuration):
        if not isinstance(scan, BlissScan):
            raise TypeError(
                "Only manage BlissScan. {} is not managed" "".format(type(scan))
            )
        assert isinstance(configuration, HDF5Config)
        self._scan = scan
        self._configuration = configuration

    def run(self):
        self.sigComputationStarted.emit()
        _logger.processStarted(
            "Start conversion of bliss scan {}".format(str(self._scan))
        )

        data_listener = DataListener()
        data_listener.set_configuration(self._configuration or {})
        try:
            scans = data_listener.process_sample_file(
                sample_file=self._scan.master_file,
                entry=self._scan.entry,
                proposal_file=self._scan.proposal_file,
                master_sample_file=self._scan.saving_file,
            )
        except Exception as e:
            _logger.processFailed(
                "Conversion of bliss scan {}." "Reason is {}".format(str(self._scan), e)
            )
            scans = None
        else:
            _logger.processSucceed(
                "Conversion of bliss scan {}." "".format(str(self._scan))
            )
        self.parent().register_result(self._scan, scans)



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################
"""Define some processing stack"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "08/01/2020"


from queue import Queue
from silx.gui import qt
from processview.core.manager import ProcessManager, DatasetState
from processview.core.superviseprocess import SuperviseProcess


class ProcessingThread(qt.QThread):
    """Class for running some processing"""

    sigComputationStarted = qt.Signal()
    """Signal emitted when a computation is started"""


class FIFO(Queue, SuperviseProcess):
    """Processing Queue with a First In, First Out behavior"""

    sigComputationStarted = qt.Signal(object)
    """Signal emitted when a computation is started"""
    sigComputationEnded = qt.Signal(object, object)
    """Signal emitted when a computation is ended. First parameter is mandatory and
    is the scan computed. Second one is an optional FutureTomwerScan"""

    def __init__(self, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        Queue.__init__(self)
        self._computationThread = self._create_processing_thread(process_id=process_id)
        assert isinstance(self._computationThread, ProcessingThread)
        self._computationThread.sigComputationStarted.connect(
            self._start_threaded_computation
        )
        """scan process by the thread"""
        self._scan_currently_computed = None
        """Scan computed currently"""
        self._processing = False

    def add(self, scan, configuration=None, callback=None):
        """
        add a scan to process

        :param TomwerScanBase scan: scan to process
        :param configuration: configuration of the process
        :param callback: function to call once the processing is Done
        """
        try:
            if self.process_id not in (None, -1):
                ProcessManager().notify_dataset_state(
                    dataset=scan,
                    process=self,
                    state=DatasetState.PENDING,
                )
        except Exception:
            pass
        Queue.put(self, (scan, configuration, callback))
        if self.can_process_next():
            self._process_next()

    def _process(self, scan, configuration, callback):
        raise NotImplementedError("Virtual class")

    def _process_next(self):
        if Queue.empty(self):
            return

        self._processing = True
        scan, configuration, callback = Queue.get(self)
        self._process(scan=scan, configuration=configuration or {}, callback=callback)

    def can_process_next(self):
        """
        :return: True if the computation thread is ready to compute
        a new axis position
        :rtype: bool
        """
        return not self._processing

    def _end_computation(self, scan, future_scan, callback):
        """
        callback when the computation thread is finished

        :param scan: pass if no call to '_computationThread is made'
        """
        if callback is not None:
            callback()
        sender = self.sender()
        if hasattr(sender, "future_scan"):
            future_scan = sender.future_scan
        else:
            future_scan = None

        self.sigComputationEnded.emit(scan, future_scan)
        self._processing = False
        if self.can_process_next():
            self._process_next()

    def _create_processing_thread(self, process_id=None) -> ProcessingThread:
        raise NotImplementedError("Virtual class")

    def is_computing(self):
        """Return True if processing thread is running (mean that computation
        is on going)"""
        return self._processing

    def wait_computation_finished(self):
        """
        Wait until the computation is finished
        """
        if self._processing:
            self._computationThread.wait()

    def _start_threaded_computation(self, *args, **kwargs):
        self.sigComputationStarted.emit(self._scan_currently_computed)

    def _end_threaded_computation(self, callback=None):
        sender = self.sender()
        if hasattr(sender, "future_scan"):
            future_scan = sender.future_scan
        else:
            future_scan = None
        self._end_computation(
            scan=self._scan_currently_computed,
            callback=callback,
            future_scan=future_scan,
        )

    def stop(self):
        while not self.empty():
            self.get()
        self._computationThread.blockSignals(True)
        self._computationThread.wait()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/03/2019"


from .base import _ReconsParam
from ....utils import _assert_cast_to_boolean, _assert_param_instance
from collections.abc import Iterable
import logging

_logger = logging.getLogger(__name__)


class PyhstRP(_ReconsParam):

    OFFV = "pyhst2"

    def __init__(self):
        _ReconsParam.__init__(self)
        self.__offv = PyhstRP.OFFV
        # textread([self.DIR self.OFFN], '%s'){1} # ne dois jamais changer On ne la presente pas
        self.__pyhst_exe = PyhstRP.OFFV
        self.__verbose_file = "pyhst_out.txt"
        self.__verbose = False
        self.__make_oar_file = False
        self._cuda_devices = None
        """Cuda gpu devices to use for the reconstruction. For now limited on
        local gpu device"""

        self._managed_params = {
            "OFFV": self.__class__._offv,
            "EXE": self.__class__.pyhst_exe,
            "VERBOSE_FILE": self.__class__.verbose_file,
            "VERBOSE": self.__class__.verbose,
            "MAKE_OAR_FILE": self.__class__.make_oar_file,
            "CUDA_DEVICES": self.__class__.cuda_devices,
        }

    @property
    def offv(self):
        return self.__offv

    @offv.setter
    def offv(self, offv):
        if self.__offv != offv:
            self.__offv = offv
            self.changed()

    @property
    def pyhst_exe(self):
        """name of the pyhste executable"""
        return self.__pyhst_exe

    @pyhst_exe.setter
    def pyhst_exe(self, exe_name):
        if self.__pyhst_exe != exe_name:
            self.__pyhst_exe = exe_name
            self.changed()

    @property
    def _offv(self):
        return self.__offv

    @_offv.setter
    def _offv(self, offv):
        if self.__offv != offv:
            self.__offv = offv
            self.changed()

    @property
    def verbose_file(self):
        """output file name if verbose is activated"""
        return self.__verbose_file

    @verbose_file.setter
    def verbose_file(self, file_name):
        if self.__verbose_file != file_name:
            self.__verbose_file = file_name
            self.changed()

    @property
    def verbose(self):
        return self.__verbose

    @verbose.setter
    def verbose(self, activate):
        assert isinstance(activate, (bool, int, float))
        if self.__verbose != bool(activate):
            self.__verbose = bool(activate)
            self.changed()

    @property
    def make_oar_file(self):
        return self.__make_oar_file

    @make_oar_file.setter
    def make_oar_file(self, make):
        _assert_param_instance(make, (bool, int, float))
        _assert_cast_to_boolean(make)
        if self.__make_oar_file != bool(make):
            self.__make_oar_file = bool(make)
            self.changed()

    @property
    def cuda_devices(self):
        """

        :return: list of cuda devices to use
        """
        return self._cuda_devices

    @cuda_devices.setter
    def cuda_devices(self, devices):
        """

        :param Union[list,None,str] devices: list of :class:`CudaDevice`
        :return: list of :class:`CudaDevice`
        """
        if isinstance(devices, Iterable):
            if len(devices) == 0:
                self._cuda_devices = None
            else:
                from tomwer.core.utils.device import CudaDevice

                self._cuda_devices = []
                for item in devices:
                    if isinstance(item, CudaDevice):
                        self._cuda_devices.append(item.id)
                    else:
                        self._cuda_devices.append(item)
        elif devices in ("", None):
            # string used to store 'no device' state
            self._cuda_devices = None
        else:
            raise TypeError(
                "devices is expected to be None or an Iterable containing instances of CudaDevice"
            )

    def to_dict(self):
        if self._cuda_devices is not None:
            cuda_devices_id = self.cuda_devices
        else:
            cuda_devices_id = ""
        _dict = {
            "OFFV": self.offv,
            "EXE": self.pyhst_exe,
            "VERBOSE_FILE": self.verbose_file,
            "VERBOSE": self.verbose,
            "MAKE_OAR_FILE": self.make_oar_file,
            "CUDA_DEVICES": cuda_devices_id,
        }
        _dict.update(self.unmanaged_params)
        return _dict

    @staticmethod
    def from_dict(_dict):
        recons_param = PyhstRP()
        recons_param.load_from_dict(_dict)
        return recons_param

    def load_from_dict(self, _dict):
        self._load_unmanaged_params(_dict)
        self.offv = _dict["OFFV"]
        self.pyhst_exe = _dict["EXE"]
        self.verbose_file = _dict["VERBOSE_FILE"]
        self.verbose = _dict["VERBOSE"]
        self.make_oar_file = _dict["MAKE_OAR_FILE"]
        try:
            self.cuda_devices = _dict["CUDA_DEVICES"]
        except KeyError:
            _logger.info('no "CUDA_DEVICES" key')



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/01/2018"


import logging
from silx.gui import qt
from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefs
from tomwer.core.process.reconstruction.darkref.params import When
from tomwer.gui.reconstruction.darkref.darkrefwidget import _WhatCheckBox
from tomwer.gui.reconstruction.ftserie.h5editor import H5StructEditor
from tomwer.synctools.ftseries import QReconsParams, _QDKRFRP

logger = logging.getLogger(__name__)


class DKRFWidget(H5StructEditor, qt.QWidget):
    """
    Group all the variable associated to dark and flat field

     :param : reconstruction parameters edited by the widget
    """

    def __init__(self, reconsparams, parent=None):
        qt.QWidget.__init__(self, parent)
        H5StructEditor.__init__(self, structID="DKRF")
        self._recons_params = None
        self.setReconsParams(recons_params=reconsparams)

        self._updateReconsParam = True

        self.setLayout(qt.QVBoxLayout())
        self.layout().addWidget(self.__buildDoWhen())
        self.layout().addWidget(self.__buildDKMode())
        self.layout().addWidget(self.__buildRefMode())
        self.layout().addWidget(self.__buildRm())
        self.layout().addWidget(self.__buildSkip())
        self.layout().addWidget(self.__buildDKPattern())
        self.layout().addWidget(self.__buildRefPattern())

        # spacer
        self.spacer = qt.QWidget(self)
        self.spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(self.spacer)

        self._makeConnection()

    def setReconsParams(self, recons_params):
        if isinstance(recons_params, QReconsParams):
            _recons_params = recons_params.dkrf
        elif isinstance(recons_params, _QDKRFRP):
            _recons_params = recons_params
        else:
            raise ValueError(
                "recons_params should be an instance of QReconsParam or _QDKRFRP"
            )

        if self._recons_params:
            self._recons_params.sigChanged.disconnect(self._update_params)
        self._recons_params = _recons_params
        self.load(self._recons_params)
        self._recons_params.sigChanged.connect(self._update_params)

    def _update_params(self):
        """Update all parameter"""
        self.load(self._recons_params)

    def _makeConnection(self):
        self._qcbRefMode.sigChanged.connect(self._refCalcModeChanged)
        self._qcbDKMode.sigChanged.connect(self._darkCalcModeChanged)
        self._qcbRmRef.toggled.connect(self._rmOptChanged)
        self._qcbSkipRef.toggled.connect(self._skipOptChanged)
        self._qleDKPattern.editingFinished.connect(self._darkPatternChanged)
        self._qleRefsPattern.editingFinished.connect(self._refPatternChanged)

    def __buildDoWhen(self):
        self._doWhenOpt = qt.QWidget(self)
        self._doWhenOpt.setLayout(qt.QHBoxLayout())
        self._doWhenOpt.layout().addWidget(qt.QLabel("when", parent=self._doWhenOpt))
        self._qleDoWhen = qt.QLineEdit("", self._doWhenOpt)
        self._doWhenOpt.layout().addWidget(self._qleDoWhen)
        self.linkGroupWithH5Variable(
            self._qleDoWhen, "DOWHEN", setter=self._setWhen, getter=self._getWhen
        )
        self._doWhenOpt.hide()
        return self._doWhenOpt

    def _setWhen(self, when):
        if isinstance(when, str):
            self._qleDoWhen.setText(when.split(".")[-1])
        elif isinstance(when, When):
            self._qleDoWhen.setText(when.name.split(".")[-1])

    def _getWhen(self):
        return self._qleDoWhen.text()

    def __buildDKMode(self):
        self._qcbDKMode = _WhatCheckBox(parent=self, text="dark")
        self.linkGroupWithH5Variable(
            self._qcbDKMode,
            "DARKCAL",
            getter=self._qcbDKMode.getModeName,
            setter=self._qcbDKMode.setMode,
        )
        return self._qcbDKMode

    def _darkCalcModeChanged(self):
        value = self._qcbDKMode.getMode()
        self._recons_params["DARKCAL"] = value

    def __buildRefMode(self):
        self._qcbRefMode = _WhatCheckBox(parent=self, text="ref")
        self.linkGroupWithH5Variable(
            self._qcbRefMode,
            "REFSCAL",
            getter=self._qcbRefMode.getModeName,
            setter=self._qcbRefMode.setMode,
        )
        return self._qcbRefMode

    def _refCalcModeChanged(self):
        value = self._qcbRefMode.getMode()
        self._recons_params["REFSCAL"] = value

    def __buildRm(self):
        widget = qt.QWidget(parent=self)
        widget.setLayout(qt.QVBoxLayout())
        widget.layout().setContentsMargins(0, 0, 0, 0)
        self._qcbRmRef = qt.QCheckBox("remove", parent=widget)
        widget.layout().addWidget(self._qcbRmRef)
        self.linkCheckboxWithH5Variable(self._qcbRmRef, "REFSRMV")

        self._qcbRmDK = qt.QCheckBox("remove dark", parent=widget)
        self._qcbRmDK.hide()
        self.linkCheckboxWithH5Variable(self._qcbRmDK, "DARKRMV")
        widget.layout().addWidget(self._qcbRmDK)

        return widget

    def _rmOptChanged(self, b):
        self._recons_params["REFSRMV"] = b
        self._recons_params["DARKRMV"] = b

    def __buildSkip(self):
        widget = qt.QWidget(parent=self)
        widget.setLayout(qt.QVBoxLayout())
        widget.layout().setContentsMargins(0, 0, 0, 0)
        self._qcbSkipRef = qt.QCheckBox("skip", parent=widget)
        widget.layout().addWidget(self._qcbSkipRef)
        self.linkCheckboxWithH5Variable(self._qcbSkipRef, "REFSOVE", invert=True)

        self._qcbSkipDK = qt.QCheckBox("skip dark", parent=widget)
        self._qcbSkipDK.hide()
        self.linkCheckboxWithH5Variable(self._qcbSkipDK, "DARKOVE", invert=True)
        widget.layout().addWidget(self._qcbSkipDK)

        return widget

    def _skipOptChanged(self, b):
        # The interface control both REFSRMV and DARKRMV
        self._recons_params._set_skip_if_exist(b)

    def __buildDKPattern(self):
        self._DKPatternOpt = qt.QWidget(self)
        self._DKPatternOpt.setLayout(qt.QHBoxLayout())
        self._DKPatternOpt.layout().addWidget(
            qt.QLabel("dark file pattern", parent=self._DKPatternOpt)
        )
        self._qleDKPattern = qt.QLineEdit("", self._DKPatternOpt)
        self._DKPatternOpt.layout().addWidget(self._qleDKPattern)
        self.LinkLineEditWithH5Variable(self._qleDKPattern, "DKFILE", str)
        self._qleDKPattern.setToolTip(DarkRefs.getDarkPatternTooltip())
        return self._DKPatternOpt

    def _darkPatternChanged(self):
        value = self._qleDKPattern.text()
        self._recons_params["DKFILE"] = value

    def __buildRefPattern(self):
        self._RefsPatternOpt = qt.QWidget(self)
        self._RefsPatternOpt.setLayout(qt.QHBoxLayout())
        self._RefsPatternOpt.layout().addWidget(
            qt.QLabel("refs file pattern", parent=self._RefsPatternOpt)
        )
        self._qleRefsPattern = qt.QLineEdit("", self._RefsPatternOpt)
        self._RefsPatternOpt.layout().addWidget(self._qleRefsPattern)
        self.LinkLineEditWithH5Variable(self._qleRefsPattern, "RFFILE", str)
        self._qleRefsPattern.setToolTip(DarkRefs.getRefPatternTooltip())
        return self._RefsPatternOpt

    def _refPatternChanged(self):
        value = self._qleRefsPattern.text()
        self._recons_params["RFFILE"] = value



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
contains gui relative to semi-automatic axis calculation
"""


__authors__ = [
    "H. Payno",
]
__license__ = "MIT"
__date__ = "01/02/2021"


from silx.gui import qt
from tomwer.gui import icons
from tomwer.gui.reconstruction.nabu.slices import NabuWindow
from tomwer.gui.reconstruction.axis.radioaxis import _CalculationWidget
from tomwer.synctools.axis import QAxisRP
from tomwer.core.process.reconstruction.nabu.utils import _NabuMode
from tomwer.core.process.reconstruction.saaxis.params import SAAxisParams
from tomwer.gui.reconstruction.scores.scoreplot import ScorePlot as _ScorePlot
from tomwer.gui.reconstruction.scores.scoreplot import CorSelection
from tomwer.gui.reconstruction.saaxis.corrangeselector import SliceAndCorWidget
from tomwer.gui.utils.buttons import TabBrowsersButtons
from tomwer.gui.reconstruction.scores.control import ControlWidget
from tomwer.core.process.reconstruction.saaxis.saaxis import SAAxisProcess
from tomwer.core.process.reconstruction.axis import mode as axis_mode
import numpy
from typing import Union
import logging


_logger = logging.getLogger(__file__)


class ScorePlot(_ScorePlot, constructor=CorSelection):
    """Score plot dedicated to center of rotation.
    Redefine the current score value to display both the absolute
    and the relative values
    """

    def _updateScores(self):
        scan = self.__scan() if self.__scan else None
        if scan is not None:
            if scan.saaxis_params:
                scan.saaxis_params.score_method = self.getScoreMethod()
                img_width = scan.dim_1
                # update autofocus
                SAAxisProcess.autofocus(scan)
        else:
            img_width = None
        self.setVarScores(
            scores=self._scores,
            score_method=self.getScoreMethod(),
            img_width=img_width,
            update_only_scores=True,
        )

    def _applyAutofocus(self):
        scan = self.__scan() if self.__scan else None
        if scan is None:
            return
        if scan.saaxis_params:
            best_cor = scan.saaxis_params.autofocus
            if best_cor:
                self._varSlider.setVarValue(best_cor)


class NabuAutoCorDiag(qt.QDialog):
    """
    GUI to compute an estimation of the Center Of Rotation
    """

    class CalculationWidget(_CalculationWidget):
        def _modeChanged(self, *args, **kwargs):
            super()._modeChanged()
            self.getMethodLockPB().hide()

    sigRequestAutoCor = qt.Signal()

    def __init__(self, parent=None, qarixrp=None):
        assert qarixrp is not None, "An instance of QAxisRP should be provided"
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self._automatic_cor = NabuAutoCorDiag.CalculationWidget(
            parent=self,
            axis_params=qarixrp,
        )
        self._automatic_cor.getMethodLockPB().hide()

        qcb = self._automatic_cor._qcbPosition
        for mode in (axis_mode.AxisMode.manual, axis_mode.AxisMode.read):
            idx = qcb.findText(mode.value)
            if idx >= 0:
                qcb.removeItem(idx)

        self.layout().addWidget(self._automatic_cor)
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        # buttons
        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(
            self._requestAutomaticCor
        )
        self._buttons.button(qt.QDialogButtonBox.Ok).setText("compute")

    def _requestAutomaticCor(self, *args, **kwargs):
        self.sigRequestAutoCor.emit()


class _SAAxisTabWidget(qt.QTabWidget):

    sigConfigurationChanged = qt.Signal()
    """signal emitted each time the 'input' configuration changed.
    like slice to reconstruct, number of reconstruction, research width,
    nabu reconstruction parameters..."""

    def __init__(self, parent):
        qt.QTabWidget.__init__(self, parent)
        # select slice & cor range

        self._sliceAndCorWidget = SliceAndCorWidget(self)
        sinogram_icon = icons.getQIcon("sinogram")
        self.addTab(self._sliceAndCorWidget, sinogram_icon, "slice && cor range")
        # nabu reconstruction parameters
        self._nabuSettings = NabuWindow(self)
        self._nabuSettings.setConfigurationLevel(level="required")
        self._nabuSettings.hideSlicesInterface()
        nabu_icon = icons.getQIcon("nabu")
        self.addTab(self._nabuSettings, nabu_icon, "reconstruction settings")
        # results
        self._resultsViewer = ScorePlot(self, variable_name="cor")
        results_icon = icons.getQIcon("results")
        self.addTab(self._resultsViewer, results_icon, "reconstructed slices")

        # connect signal / slot
        self._nabuSettings.sigConfigChanged.connect(self._configurationChanged)
        self._sliceAndCorWidget.sigConfigurationChanged.connect(
            self._configurationChanged
        )
        self._resultsViewer.sigConfigurationChanged.connect(self._configurationChanged)
        self.sigReconstructionSliceChanged = (
            self._sliceAndCorWidget.sigReconstructionSliceChanged
        )
        self.sigAutoCorRequested = self._sliceAndCorWidget.sigAutoCorRequested
        self.sigReconstructionRangeChanged = (
            self._sliceAndCorWidget.sigReconstructionRangeChanged
        )

        # expose function API
        self.setCorScores = self._resultsViewer.setVarScores
        self.setImgWidth = self._resultsViewer.setImgWidth
        self.setVoxelSize = self._resultsViewer.setVoxelSize
        self.setVolumeSize = self._resultsViewer.setVolumeSize
        self.setCurrentCorValue = self._resultsViewer.setCurrentVarValue
        self.getCurrentCorValue = self._resultsViewer.getCurrentVarValue
        self.getScoreMethod = self._resultsViewer.getScoreMethod
        self.getEstimatedCorPosition = self._sliceAndCorWidget.getEstimatedCorPosition
        self.setEstimatedCorPosition = self._sliceAndCorWidget.setEstimatedCorPosition
        self.getNReconstruction = self._sliceAndCorWidget.getNReconstruction
        self.setNReconstruction = self._sliceAndCorWidget.setNReconstruction
        self.getResearchWidth = self._sliceAndCorWidget.getResearchWidth
        self.setResearchWidth = self._sliceAndCorWidget.setResearchWidth
        self.getReconstructionSlices = self._sliceAndCorWidget.getReconstructionSlices
        self.setReconstructionSlices = self._sliceAndCorWidget.setReconstructionSlices
        self.getReconstructionMode = self._sliceAndCorWidget.getReconstructionMode
        self.setReconstructionMode = self._sliceAndCorWidget.setReconstructionMode
        self.getFrameWidth = self._sliceAndCorWidget.getFrameWidth
        self.setFrameWidth = self._sliceAndCorWidget.setFrameWidth
        self.setNabuReconsParams = self._nabuSettings.setConfiguration
        self.getNabuReconsParams = self._nabuSettings.getConfiguration
        self.getSlicesRange = self._sliceAndCorWidget.getSlicesRange
        self.setSlicesRange = self._sliceAndCorWidget.setSlicesRange
        self.loadSinogram = self._sliceAndCorWidget.loadSinogram
        # expose signals
        self.sigStartSinogramLoad = self._sliceAndCorWidget.sigStartSinogramLoad
        self.sigEndSinogramLoad = self._sliceAndCorWidget.sigEndSinogramLoad

    def showResults(self):
        self.setCurrentWidget(self._resultsViewer)

    def _configurationChanged(self, *args, **kwargs):
        self.sigConfigurationChanged.emit()

    def lockAutoFocus(self, lock):
        self._resultsViewer.lockAutoFocus(lock=lock)

    def isAutoFocusLock(self):
        return self._resultsViewer.isAutoFocusLock()

    def hideAutoFocusButton(self):
        self._resultsViewer.hideAutoFocusButton()

    def getSinogramViewer(self):
        return self._sliceAndCorWidget._sinogramViewer

    def getCors(self, reference: str = "relative"):
        """Return cors to be computed"""
        if reference not in ("relative", "absolute"):
            raise ValueError("reference should be 'absolute' or 'relative'")
        return SAAxisParams.compute_cors(
            estimated_cor=self.getEstimatedCorPosition(reference),
            research_width=self.getResearchWidth(),
            n_reconstruction=self.getNReconstruction(),
        )

    def loadPreprocessingParams(self):
        """load reconstruction nabu if tomwer has already process this
        dataset. Not done for now"""
        return False

    def setScan(self, scan):
        self._resultsViewer.setScan(scan)
        self._nabuSettings.setScan(scan)
        if self.loadPreprocessingParams() and scan.axis_params is not None:
            self._nabuSettings.setConfiguration(scan.axis_params)
        self._sliceAndCorWidget.setScan(scan)

    def getConfiguration(self):
        nabu_config = self.getNabuReconsParams()
        enable_ht = int(self._nabuSettings.getMode() is _NabuMode.HALF_ACQ)
        nabu_config["reconstruction"]["enable_halftomo"] = enable_ht
        return {
            "research_width": self.getResearchWidth(),
            "n_reconstruction": self.getNReconstruction(),
            "slice_indexes": self.getReconstructionSlices(),
            "nabu_params": nabu_config,
            "mode": self.getReconstructionMode().value,
            "score_method": self.getScoreMethod().value,
            "estimated_cor": self.getEstimatedCorPosition(),
        }

    def setConfiguration(self, config):
        if isinstance(config, SAAxisParams):
            config = config.to_dict()
        if not isinstance(config, dict):
            raise TypeError(
                "config should be a dictionary or a SAAxisParams. "
                "Not {}".format(type(config))
            )

        research_width = config.get("research_width", None)
        if research_width is not None:
            self.setResearchWidth(research_width)
        n_reconstruction = config.get("n_reconstruction", None)
        if n_reconstruction is not None:
            self.setNReconstruction(n_reconstruction)
        estimated_cor = config.get("estimated_cor", None)
        if estimated_cor is not None:
            self.setEstimatedCorPosition(estimated_cor)
        slice_indexes = config.get("slice_indexes", None)
        if slice_indexes is not None:
            self.setReconstructionSlices(slice_indexes)
        if "nabu_params" in config:
            self.setNabuReconsParams(config["nabu_params"])
        if "mode" in config:
            self.setReconstructionMode(config["mode"])
        if "score_method" in config:
            self.setScoreMethod(config["score_method"])

    def getScoreMethod(self):
        return self._resultsViewer.getScoreMethod()

    def setScoreMethod(self, method):
        self._resultsViewer.setScoreMethod(method)

    def close(self):
        self._resultsViewer.close()
        self._resultsViewer = None
        self._sliceAndCorWidget.close()
        self._sliceAndCorWidget = None
        self._nabuSettings.close()
        self._nabuSettings = None
        super().close()

    def _stopAnimationThread(self):
        self._resultsViewer._stopAnimationThread()
        self._sliceAndCorWidget._sinogramViewer._stopAnimationThread()


class SAAxisWindow(qt.QMainWindow):
    """
    Widget used to determine half-automatically the center of rotation
    """

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)
        self._scan = None
        self._automaticCorWidget = None
        self._qaxis_rp = QAxisRP()
        self.setWindowFlags(qt.Qt.Widget)
        # central widget
        self._tabWidget = _SAAxisTabWidget(self)
        self.setCentralWidget(self._tabWidget)
        # next and previous buttons for browsing the tab widget
        self._browserButtons = TabBrowsersButtons(self)
        self._dockWidgetBrwButtons = qt.QDockWidget(self)
        self._dockWidgetBrwButtons.setWidget(self._browserButtons)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._dockWidgetBrwButtons)
        self._dockWidgetBrwButtons.setFeatures(qt.QDockWidget.DockWidgetMovable)
        # control widget (validate, compute, cor positions)
        self._saaxisControl = ControlWidget(self)
        self._dockWidgetCtrl = qt.QDockWidget(self)
        self._dockWidgetCtrl.setWidget(self._saaxisControl)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._dockWidgetCtrl)
        self._dockWidgetCtrl.setFeatures(qt.QDockWidget.DockWidgetMovable)

        # expose API
        self.setCorScores = self._tabWidget.setCorScores
        self.setImgWidth = self._tabWidget.setImgWidth
        self.setVoxelSize = self._tabWidget.setVoxelSize
        self.setVolumeSize = self._tabWidget.setVolumeSize
        self.setCurrentCorValue = self._tabWidget.setCurrentCorValue
        self.getCurrentCorValue = self._tabWidget.getCurrentCorValue
        self.getEstimatedCorPosition = self._tabWidget.getEstimatedCorPosition
        self.setEstimatedCorPosition = self._tabWidget.setEstimatedCorPosition
        self.getNReconstruction = self._tabWidget.getNReconstruction
        self.setNReconstruction = self._tabWidget.setNReconstruction
        self.getResearchWidth = self._tabWidget.getResearchWidth
        self.setResearchWidth = self._tabWidget.setResearchWidth
        self.getReconstructionSlices = self._tabWidget.getReconstructionSlices
        self.setReconstructionSlices = self._tabWidget.setReconstructionSlices
        self.getNabuReconsParams = self._tabWidget.getNabuReconsParams
        self.setNabuReconsParams = self._tabWidget.setNabuReconsParams
        self.getSlicesRange = self._tabWidget.getSlicesRange
        self.setSlicesRange = self._tabWidget.setSlicesRange
        self.getCors = self._tabWidget.getCors
        self.getMode = self._tabWidget.getReconstructionMode
        self.loadSinogram = self._tabWidget.loadSinogram
        self.isAutoFocusLock = self._tabWidget.isAutoFocusLock
        # expose signals
        self.sigValidated = self._saaxisControl.sigValidateRequest
        self.sigStartSinogramLoad = self._tabWidget.sigStartSinogramLoad
        self.sigEndSinogramLoad = self._tabWidget.sigEndSinogramLoad
        self.sigConfigurationChanged = self._tabWidget.sigConfigurationChanged

        # connect signal / slot
        self._tabWidget.sigReconstructionSliceChanged.connect(self._updateSinogramLine)
        self._tabWidget.sigAutoCorRequested.connect(self._autoCorRequested)
        self._tabWidget.sigReconstructionRangeChanged.connect(
            self._estimatedCorValueChanged
        )
        self._browserButtons.sigNextReleased.connect(self._showNextPage)
        self._browserButtons.sigPreviousReleased.connect(self._showPreviousPage)
        self._saaxisControl.sigComputationRequest.connect(self._launchReconstructions)
        self._saaxisControl.sigValidateRequest.connect(self._validate)

    def stop(self):
        self._stopAnimationThread()

    def showResults(self):
        self._tabWidget.showResults()

    def getAutomaticCorWindow(self):
        if self._automaticCorWidget is None:
            self._automaticCorWidget = NabuAutoCorDiag(self, qarixrp=self._qaxis_rp)
            self._automaticCorWidget.setWindowTitle(
                "compute estimated center of rotation"
            )
            auto_cor_icon = icons.getQIcon("a")
            self._automaticCorWidget.setWindowIcon(auto_cor_icon)
            self._automaticCorWidget.sigRequestAutoCor.connect(
                self._computeEstimatedCor
            )
        return self._automaticCorWidget

    def compute(self):
        """force compute of the current scan"""
        self._saaxisControl.sigComputationRequest.emit()

    def getConfiguration(self) -> dict:
        return self._tabWidget.getConfiguration()

    def setConfiguration(self, config: dict):
        self._tabWidget.setConfiguration(config)

    def getQAxisRP(self):
        return self._qaxis_rp

    def setScan(self, scan):
        self._scan = scan
        self._tabWidget.setScan(scan)
        self._updateSinogramLine()
        self._loadEstimatedCorFromScan(scan)

    def _loadEstimatedCorFromScan(self, scan):
        if scan.axis_params is not None:
            relative_cor = scan.axis_params.relative_cor_value
        else:
            relative_cor = None
        if relative_cor is None:
            relative_cor = scan.estimated_cor_frm_motor

        if relative_cor is not None and numpy.issubdtype(
            type(relative_cor), numpy.number
        ):
            self.setEstimatedCorPosition(relative_cor)

    def getScan(self):
        return self._scan

    def getScoreMethod(self):
        return self._tabWidget.getScoreMethod()

    def lockAutofocus(self, lock):
        self._tabWidget.lockAutoFocus(lock=lock)

    def isAutoFocusLock(self):
        return self._tabWidget.isAutoFocusLock()

    def hideAutoFocusButton(self):
        return self._tabWidget.hideAutoFocusButton()

    def _updateSinogramLine(self):
        r_slice = self.getReconstructionSlices()
        if r_slice == "middle":
            if self._scan is not None and self._scan.dim_2 is not None:
                line = self._scan.dim_2 // 2
            else:
                line = 1024
        else:
            line = list(r_slice.values())[0]
        self._tabWidget.getSinogramViewer().setLine(line)

    def setEstimatedCorPosition(self, value):
        self.setEstimatedCorPosition(value=value)

    def _autoCorRequested(self):
        window = self.getAutomaticCorWindow()
        window.activateWindow()
        window.raise_()
        window.show()

    def _computeEstimatedCor(self) -> Union[float, None]:
        """callback when calculation of a estimated cor is requested.
        Should be implemted by OrangeWidget or application"""
        raise NotImplementedError("Base class")

    def _launchReconstructions(self):
        """callback when we want to launch the reconstruction of the
        slice for n cor value"""
        raise NotImplementedError("Base class")

    def _validate(self):
        raise NotImplementedError("Base class")

    def _estimatedCorValueChanged(self):
        cors = self.getCors("absolute")
        sino_viewer = self._tabWidget._sliceAndCorWidget._sinogramViewer
        estimated_cor = self.getEstimatedCorPosition("absolute")
        if estimated_cor == "middle":
            estimated_cor = 0

        if len(cors) < 2:
            return
        elif len(cors) > 2:
            other_cors = cors[1:-1]
        else:
            other_cors = ()
        sino_viewer.setCorRange(
            min=cors[0],
            max=cors[-1],
            estimated_cor=estimated_cor,
            other_cors=other_cors,
        )

    def _showNextPage(self, *args, **kwargs):
        idx = self._tabWidget.currentIndex()
        idx += 1
        if idx < self._tabWidget.count():
            self._tabWidget.setCurrentIndex(idx)

    def _showPreviousPage(self, *args, **kwargs):
        idx = self._tabWidget.currentIndex()
        idx -= 1
        if idx >= 0:
            self._tabWidget.setCurrentIndex(idx)

    def _stopAnimationThread(self):
        self._tabWidget._stopAnimationThread()

    def close(self):
        self._tabWidget.close()
        self._tabWidget = None
        super().close()


def _get_circle_plot(center, radius, n_pts=1000):
    pts = numpy.linspace(-numpy.pi, numpy.pi, n_pts)
    x = radius * numpy.cos(pts) + center[0]
    y = radius * numpy.sin(pts) + center[1]
    return x, y
